import errno
import os
import re
import sys
from pathlib import Path
from typing import Optional

from album.runner import album_logging

from album.ci.utils.zenodo_api import ZenodoAPI
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import (
    force_remove,
    create_path_recursively,
    rand_folder_name,
    check_zip,
    unzip_archive,
    copy,
    copy_folder,
)
from album.core.utils.operations.url_operations import (
    is_url,
    download,
    retrieve_redirect_url,
    download_resource,
)
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.model.coordinates import Coordinates

module_logger = album_logging.get_active_logger


def clean_resolve_tmp(tmp_cache_dir) -> None:
    """Cleans the temporary directory which might have been used during resolving."""
    force_remove(tmp_cache_dir)
    create_path_recursively(tmp_cache_dir)


def get_attributes_from_string(str_input: str) -> dict:
    """Interprets a string input if in valid format and returns necessary attributes dictionary.

    Args:
        str_input:
            The string input. Supported formats:
                doi:  <doi>:<prefix>/<suffix> or <prefix>/<suffix> of a solution
                gnv: <group>:<name>:<version> of a solution
                cgnv: <catalog>:<group>:<name>:<version> of a solution
                url: any url pointing to a solution file

    Returns:
        Dictionary of either grp name version, catalog grp name version or doi

    """
    attrs_dict = get_doi_from_input(str_input)
    if not attrs_dict:
        attrs_dict = get_gnv_from_input(str_input)
        if not attrs_dict:
            attrs_dict = get_cgnv_from_input(str_input)
            if not attrs_dict:
                raise ValueError(
                    "Invalid input format: %s! Try <doi>:<prefix>/<suffix> or <prefix>/<suffix> "
                    "or <group>:<name>:<version> or <catalog>:<group>:<name>:<version> "
                    "or point to a valid file! Aborting..." % str_input
                )
    module_logger().debug("Parsed %s from the input... " % attrs_dict)
    return attrs_dict


def get_gnv_from_input(str_input: str):
    """Parses Group, Name, Version from input, separated by ":"."""
    s = re.search("^([^:]+):([^:]+):([^:]+)$", str_input)

    if s:
        return {"group": s.group(1), "name": s.group(2), "version": s.group(3)}
    return None


def get_cgnv_from_input(str_input: str):
    """Parses Catalog, Group, Name, Version from input, separated by ":"."""
    s = re.search("^([^:]+):([^:]+):([^:]+):([^:]+)$", str_input)

    if s:
        return {
            "catalog": s.group(1),
            "group": s.group(2),
            "name": s.group(3),
            "version": s.group(4),
        }
    return None


def get_doi_from_input(str_input: str):
    """Parses the DOI from string input."""
    s = re.search("^(^doi:)?([^:\/]*\/[^:\/]*)$", str_input)
    if s:
        return {"doi": s.group(2)}
    return None


def is_pathname_valid(pathname: str) -> bool:
    """Checks if a pathname is valid for the current OS

    `True` if the passed pathname is a valid pathname for the current OS;
    `False` otherwise.
    source: https://stackoverflow.com/a/34102855
    """
    # Windows-specific error code indicating an invalid pathname.
    # Sadly, Python fails to provide the following magic number for us.
    ERROR_INVALID_NAME = 123
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = (
            os.environ.get("HOMEDRIVE", "C:")
            if sys.platform == "win32"
            else os.path.sep
        )
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, "winerror"):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.


def check_doi(doi, tmp_cache_dir):
    tmp_cache_dir = Path(tmp_cache_dir).joinpath(rand_folder_name())

    link = "https://doi.org/" + doi  # e.g. 10.5281/zenodo.5571504

    url = retrieve_redirect_url(link)

    link_to_solution_zip = parse_doi_service_url(url)

    p = download_resource(link_to_solution_zip, tmp_cache_dir.joinpath("solution.zip"))

    return prepare_path(p, tmp_cache_dir)


def parse_doi_service_url(url):
    if re.search("https:\/\/[a-zA-Z.]*zenodo[.]org\/", url):
        link = _parse_zenodo_url(url)
    else:
        raise NotImplementedError("DOI service not supported!")

    return link


def _parse_zenodo_url(url):
    g = re.search("(https:\/\/[a-zA-Z.]*zenodo[.]org\/)(record)[\/]([0-9]*)$", url)

    if g:
        record_id = g.group(3)

        return retrieve_zenodo_record_download_zip(record_id)
    else:
        raise ValueError("Unknown zenodo URL format!...")


def retrieve_zenodo_record_download_zip(record_id):
    query = ZenodoAPI()

    record = query.records_get(record_id)[0]

    file_dl = None
    for file in record.files:
        if re.search(
            "[.]zip$", file.key
        ):  # there is only a single zip file in the record
            file_dl = file.get_download_link()

    if not file_dl:
        raise ValueError(
            'No valid zip file found int the zenodo record with id "%s"!' % record_id
        )

    return file_dl


def check_file_or_url(path, tmp_cache_dir):
    """Resolves a path or url. Independent of catalogs."""
    if is_url(path):
        p = download(str(path), base=tmp_cache_dir)
    elif is_pathname_valid(path) and (os.path.isfile(path) or os.path.isdir(path)):
        p = Path(path)
    else:
        return None

    return prepare_path(p, tmp_cache_dir)


def prepare_path(path, tmp_cache_dir):
    """Prepares the path to be run in album. Returning path points to the solution file.

    Args:
        path:
            The path pointing to a zip, file, or folder suited to be run in album.
        tmp_cache_dir:
            The temporary cache dir to extract/copy files to.

    Returns:
        The path pointing to the python file executable in album.

    """
    p = Path(path)
    if p.exists():
        # copying to tmp-dir necessary for cloning
        target_folder = tmp_cache_dir.joinpath(rand_folder_name())
        if p.is_file():  # zip or file
            if check_zip(p):  # zip file
                p = unzip_archive(p, target_folder)
                p = p.joinpath(DefaultValues.solution_default_name.value)
            else:  # python file
                p = copy(
                    p, target_folder.joinpath(DefaultValues.solution_default_name.value)
                )
        elif p.is_dir():  # unzipped zip
            p = copy_folder(p, target_folder, copy_root_folder=False)
            p = p.joinpath(DefaultValues.solution_default_name.value)

        return p


def dict_to_coordinates(solution_attr) -> ICoordinates:
    if not all([k in solution_attr.keys() for k in ["name", "version", "group"]]):
        raise ValueError(
            "Cannot resolve solution! Group, name and version must be specified!"
        )
    return Coordinates(
        group=solution_attr["group"],
        name=solution_attr["name"],
        version=solution_attr["version"],
    )


def get_zip_name(coordinates: ICoordinates):
    return get_zip_name_prefix(coordinates) + ".zip"


def get_zip_name_prefix(coordinates: ICoordinates):
    return "_".join([coordinates.group(), coordinates.name(), coordinates.version()])


def build_resolve_string(
    resolve_solution_dict: dict, catalog: Optional[ICatalog] = None
):
    if "doi" in resolve_solution_dict.keys():
        resolve_solution = resolve_solution_dict["doi"]
    elif all([x in resolve_solution_dict.keys() for x in ["group", "name", "version"]]):
        resolve_solution = ":".join(
            [
                resolve_solution_dict["group"],
                resolve_solution_dict["name"],
                resolve_solution_dict["version"],
            ]
        )

        if catalog:
            resolve_solution = ":".join([catalog.name(), resolve_solution])

    elif "resolve_solution" in resolve_solution_dict.keys():
        resolve_solution = resolve_solution_dict["resolve_solution"]
    else:
        raise ValueError("Invalid declaration of parent or step!")

    return resolve_solution


def get_parent(
    parent_collection_entry: ICollectionIndex.ICollectionSolution,
) -> ICollectionIndex.ICollectionSolution:
    """Given an collection entry (aka row of the collection table) this method returns the corresponding parent"""
    if parent_collection_entry.internal()["parent"]:
        parent = parent_collection_entry.internal()["parent"]
        while parent.internal()["parent"]:
            parent = parent["parent"]

        return parent
    return parent_collection_entry


def as_tag(coordinates: ICoordinates) -> str:
    return "-".join([coordinates.group(), coordinates.name(), coordinates.version()])


def as_tag_unversioned(coordinates: ICoordinates) -> str:
    return "-".join([coordinates.group(), coordinates.name()])



from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.environment_manager import IEnvironmentManager
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.api.model.environment import IEnvironment
from album.core.controller.conda_manager import CondaManager
from album.core.controller.mamba_manager import MambaManager
from album.core.model.environment import Environment
from album.core.utils.operations.file_operations import remove_link
from album.core.utils.operations.resolve_operations import dict_to_coordinates
from album.core.utils.operations.solution_operations import set_environment_paths
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates

module_logger = album_logging.get_active_logger


class EnvironmentManager(IEnvironmentManager):
    def __init__(self, album: IAlbumController):
        self._conda_manager = CondaManager(album.configuration())
        if album.configuration().mamba_executable():
            self._env_install_manager = MambaManager(album.configuration())
        else:
            self._env_install_manager = self._conda_manager
        self._album = album

    def install_environment(
        self, collection_solution: ICollectionSolution
    ) -> IEnvironment:
        environment = Environment(
            collection_solution.loaded_solution().setup().dependencies,
            self.get_environment_name(
                collection_solution.coordinates(), collection_solution.catalog()
            ),
            collection_solution.loaded_solution().installation().internal_cache_path(),
        )
        self._env_install_manager.install(
            environment, collection_solution.loaded_solution().setup().album_api_version
        )
        set_environment_paths(collection_solution.loaded_solution(), environment)
        return environment

    def set_environment(self, collection_solution: ICollectionSolution) -> IEnvironment:
        parent = collection_solution.database_entry().internal()["parent"]
        # solution runs in its own environment
        if not parent:

            environment = Environment(
                dependencies_dict=None,
                environment_name=self.get_environment_name(
                    collection_solution.coordinates(), collection_solution.catalog()
                ),
                cache_path=collection_solution.loaded_solution()
                .installation()
                .internal_cache_path(),
            )
            self._conda_manager.set_environment_path(environment)

        # solution runs in the parents environment - we need to resolve first to get info about parents environment
        else:
            coordinates = dict_to_coordinates(parent.setup())
            catalog = self._album.catalogs().get_by_id(parent.internal()["catalog_id"])

            environment = Environment(
                None,
                self.get_environment_name(coordinates, catalog),
                collection_solution.loaded_solution()
                .installation()
                .internal_cache_path(),
            )
            self._conda_manager.set_environment_path(environment)

        set_environment_paths(collection_solution.loaded_solution(), environment)
        return environment

    def remove_environment(self, environment: IEnvironment) -> bool:
        """Removes an environment."""
        res = self._conda_manager.remove_environment(environment.name())
        self.remove_disc_content_from_environment(environment)
        return res

    def run_scripts(self, environment: IEnvironment, scripts, pipe_output=True):
        if environment:
            self._conda_manager.run_scripts(
                environment, scripts, pipe_output=pipe_output
            )
        else:
            raise EnvironmentError("Environment not set! Cannot run scripts!")

    def get_conda_manager(self):
        return self._conda_manager

    @staticmethod
    def get_environment_name(coordinates: ICoordinates, catalog: ICatalog) -> str:
        return "_".join(
            [
                str(catalog.name()),
                coordinates.group(),
                coordinates.name(),
                coordinates.version(),
            ]
        )

    @staticmethod
    def remove_disc_content_from_environment(environment: IEnvironment):
        remove_link(environment.path())
        remove_link(environment.cache_path())



from pathlib import Path
from urllib.parse import urlparse

from album.core.utils.operations.file_operations import write_dict_to_yml

from album.runner import album_logging

module_logger = album_logging.get_active_logger


def get_ssh_url(project_path, server_http_url):
    parsed_url = urlparse(server_http_url)

    ssh_url = "git@%s:%s" % (parsed_url.netloc, project_path)

    module_logger().debug("Set remote URL to %s..." % ssh_url)

    return ssh_url


def create_report(report_file: Path, report_vars: dict) -> Path:
    write_dict_to_yml(report_file, report_vars)

    return report_file



from abc import ABCMeta, abstractmethod


class IRunManager:
    """Interface managing the running process of a solution.

    A solution is executed in its target environment which is created during installation. This class performs all
    operations necessary to run a solution. Resolving of a solution in all configured catalogs,
    dependency checking, and more.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def run(self, solution_to_resolve: str, run_immediately=False, argv=None):
        """Function corresponding to the `run` subcommand of `album`."""
        raise NotImplementedError



import copy
import hashlib
import json
from datetime import date, time
from typing import Optional

from album.core.api.model.collection_solution import ICollectionSolution
from album.core.api.model.environment import IEnvironment
from album.core.utils.operations.file_operations import remove_link
from album.runner import album_logging
from album.runner.core.api.model.solution import ISolution

module_logger = album_logging.get_active_logger


def set_environment_paths(solution: ISolution, environment: IEnvironment):
    """Sets the available cache paths of the solution object, given the environment used to run it."""
    solution.installation().set_environment_path(environment.path())
    solution.installation().set_environment_name(environment.name())


def get_deploy_dict(solution: ISolution) -> dict:
    """Return a dictionary with the relevant deployment key/values for a given album."""
    d = {}

    for k in solution.setup().keys():
        value = solution.setup()[k]
        if not callable(value) and k != "dependencies":
            # deepcopy necessary. Else original album object will loose "action" attributes in its arguments
            d[k] = copy.deepcopy(value)

    return _remove_action_from_args(d)


def _remove_action_from_args(solution_dict):
    if "args" in solution_dict:
        for arg in solution_dict["args"]:
            if isinstance(arg, dict):
                # iterate and remove callable values
                for key in arg.keys():
                    if callable(arg[key]):
                        arg[key] = "%s_function" % key
    return solution_dict


def get_parent_dict(solution: ISolution) -> Optional[dict]:
    if solution.setup().dependencies and "parent" in solution.setup().dependencies:
        return solution.setup().dependencies["parent"]
    return None


def get_steps_dict(solution: ISolution) -> Optional[dict]:
    return solution.setup().steps


def create_hash(string_representation):
    hash_val = hashlib.md5(string_representation.encode("utf-8")).hexdigest()

    return hash_val


def get_solution_hash(solution_attrs, keys):
    return create_hash(
        ":".join([json.dumps(solution_attrs[k]) for k in keys if k in solution_attrs])
    )


def serialize_json(catalogs_as_dict):
    return json.dumps(catalogs_as_dict, sort_keys=True, indent=4, default=serialize)


def serialize(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, date):
        serial = obj.isoformat()
        return serial

    if isinstance(obj, time):
        serial = obj.isoformat()
        return serial

    return obj.__dict__



import gc
import logging
import sys
import tempfile
import unittest
from io import StringIO
from pathlib import Path
from typing import Optional
from unittest.mock import patch

from album.api import Album
from album.core.controller.album_controller import AlbumController
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import write_dict_to_json, force_remove
from album.core.utils.operations.git_operations import (
    create_bare_repository,
    clone_repository,
    add_files_commit_and_push,
)
from album.core.utils.operations.view_operations import (
    get_message_filter,
    get_logger_name_minimizer_filter,
    get_logging_formatter,
)
from album.runner.album_logging import get_active_logger
from test.global_exception_watcher import GlobalExceptionWatcher


class TestCommon(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.setup_tmp_resources()
        self.setup_silent_test_logging()
        self.album: Optional[Album] = None
        self.album_controller: Optional[AlbumController] = None

    def tearDown(self) -> None:
        if self.album:
            self.album.close()
            self.album_patch.stop()
        self.logger.handlers.clear()
        self.teardown_tmp_resources()
        super().tearDown()

    def setup_tmp_resources(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.closed_tmp_file = tempfile.NamedTemporaryFile(delete=False)
        self.closed_tmp_file.close()

    def setup_album_controller(self):
        if self.album_controller:
            raise AttributeError(
                "Only one instance of an AlbumController should be used!"
                "Either use setup_album_instance or setup_album_controller!"
            )
        self.album_controller = AlbumController(base_cache_path=Path(self.tmp_dir.name))

    def setup_album_instance(self):
        if self.album:
            raise AttributeError(
                "Only one instance of an Album should be used!"
                "Either use setup_album_instance or setup_album_controller!"
            )
        with patch("album.api.configure_root_logger"):
            self.album = (
                Album.Builder().base_cache_path(Path(self.tmp_dir.name)).build()
            )
        if self.album_controller:
            self.album._controller = self.album_controller
        else:
            self.album_controller = self.album._controller
        self.album_patch = patch(
            "album.argument_parsing.create_album_instance", return_value=self.album
        )
        self.album_patch.start()

    def setup_silent_test_logging(self, logger_name="test-logger"):
        self.captured_output = StringIO()
        self.logger = get_active_logger()
        self.logger.handlers.clear()
        self.logger.name = logger_name
        self.logger.setLevel("INFO")
        ch = logging.StreamHandler(self.captured_output)
        ch.setLevel("INFO")
        ch.setFormatter(
            get_logging_formatter(
                fmt="%(log_color)s%(asctime)s %(levelname)s %(shortened_name)s%(message)s"
            )
        )
        ch.addFilter(get_logger_name_minimizer_filter())
        ch.addFilter(get_message_filter())
        self.logger.addHandler(ch)
        return self.logger

    def setup_empty_catalog(self, name, catalog_type="direct"):
        catalog_src_path = Path(self.tmp_dir.name).joinpath("my-catalogs", name)
        create_bare_repository(catalog_src_path)

        catalog_clone_path = Path(self.tmp_dir.name).joinpath("my-catalogs-clone", name)

        with clone_repository(catalog_src_path, catalog_clone_path) as repo:
            head = repo.active_branch

            self.album_controller.catalogs().create_new_metadata(
                catalog_clone_path, name, catalog_type
            )

            add_files_commit_and_push(
                head,
                [catalog_clone_path],
                "init",
                push=True,
                username=DefaultValues.catalog_git_user.value,
                email=DefaultValues.catalog_git_email.value,
            )

        return catalog_src_path, catalog_clone_path

    def setup_collection(self, init_catalogs=True, init_collection=True):
        if init_catalogs:
            with patch(
                "album.core.model.configuration.Configuration.get_initial_catalogs"
            ) as get_initial_catalogs_mock:
                get_initial_catalogs_mock.return_value = {}

                # create collection
                self.album_controller.collection_manager().load_or_create()
        elif init_collection:
            with patch(
                "album.core.controller.collection.catalog_handler.CatalogHandler.add_initial_catalogs"
            ):
                self.album_controller.collection_manager().load_or_create()
        # check everything is freshly initialized
        self.assertTrue(
            self.album_controller.collection_manager().get_collection_index().is_empty()
        )

    def teardown_tmp_resources(self):
        # garbage collector
        gc.collect()

        try:
            Path(self.closed_tmp_file.name).unlink()
            self.tmp_dir.cleanup()
        except PermissionError:
            try:
                force_remove(self.tmp_dir.name)
            except PermissionError:
                if sys.platform == "win32" or sys.platform == "cygwin":
                    pass
                else:
                    raise

    def teardown_logging(self):
        self.logger.handlers.clear()
        if self.album:
            album_logger = logging.getLogger("album")
            album_logger.handlers.clear()

    def get_logs(self):
        logs = self.captured_output.getvalue()
        logs = logs.strip()
        return logs.split("\n")

    def run(self, result=None):
        # add watcher to catch any exceptions thrown in threads
        with GlobalExceptionWatcher():
            super(TestCommon, self).run(result)

    @staticmethod
    def get_catalog_meta_dict(
        name="cache_catalog", version="0.1.0", catalog_type="direct"
    ):
        return {"name": name, "version": version, "type": catalog_type}



from pathlib import Path
from typing import Optional

from album.core.api.controller.collection.catalog_handler import ICatalogHandler
from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.environment_manager import IEnvironmentManager
from album.core.api.controller.event_manager import IEventManager
from album.core.api.controller.migration_manager import IMigrationManager
from album.core.api.controller.script_manager import IScriptManager
from album.core.api.controller.task_manager import ITaskManager
from album.core.api.model.configuration import IConfiguration
from album.core.controller.clone_manager import CloneManager
from album.core.controller.collection.collection_manager import CollectionManager
from album.core.controller.deploy_manager import DeployManager
from album.core.controller.environment_manager import EnvironmentManager
from album.core.controller.event_manager import EventManager
from album.core.controller.install_manager import InstallManager
from album.core.controller.migration_manager import MigrationManager
from album.core.controller.run_manager import RunManager
from album.core.controller.script_manager import ScriptManager
from album.core.controller.search_manager import SearchManager
from album.core.controller.state_manager import StateManager
from album.core.controller.task_manager import TaskManager
from album.core.controller.test_manager import TestManager
from album.core.model.configuration import Configuration


class AlbumController(IAlbumController):
    def __init__(self, base_cache_path: Optional[Path] = None) -> None:
        self.base_cache_path = base_cache_path
        self._deploy_manager = None
        self._run_manager = None
        self._install_manager = None
        self._collection_manager: Optional[CollectionManager] = None
        self._test_manager = None
        self._migration_manager = None
        self._search_manager = None
        self._clone_manager = None
        self._environment_manager = None
        self._state_manager = None
        self._script_manager = None
        self._event_manager = None
        self._task_manager = None
        self._configuration = None

    def catalogs(self) -> ICatalogHandler:
        return self.collection_manager().catalogs()

    def solutions(self) -> ISolutionHandler:
        return self.collection_manager().solutions()

    def configuration(self) -> IConfiguration:
        if not self._configuration:
            self._configuration = Configuration()
            self._configuration.setup(base_cache_path=self.base_cache_path)
        return self._configuration

    def environment_manager(self) -> IEnvironmentManager:
        if not self._environment_manager:
            self._environment_manager = EnvironmentManager(self)
        return self._environment_manager

    def migration_manager(self) -> IMigrationManager:
        if not self._migration_manager:
            self._migration_manager = MigrationManager(self)
        return self._migration_manager

    def script_manager(self) -> IScriptManager:
        if not self._script_manager:
            self._script_manager = ScriptManager(self)
        return self._script_manager

    def deploy_manager(self) -> DeployManager:
        if not self._deploy_manager:
            self._deploy_manager = DeployManager(self)
        return self._deploy_manager

    def install_manager(self) -> InstallManager:
        if not self._install_manager:
            self._install_manager = InstallManager(self)
        return self._install_manager

    def run_manager(self) -> RunManager:
        if not self._run_manager:
            self._run_manager = RunManager(self)
        return self._run_manager

    def test_manager(self) -> TestManager:
        if not self._test_manager:
            self._test_manager = TestManager(self)
        return self._test_manager

    def search_manager(self) -> SearchManager:
        if not self._search_manager:
            self._search_manager = SearchManager(self)
        return self._search_manager

    def clone_manager(self) -> CloneManager:
        if not self._clone_manager:
            self._clone_manager = CloneManager(self)
        return self._clone_manager

    def state_manager(self) -> StateManager:
        if not self._state_manager:
            self._state_manager = StateManager(self)
        return self._state_manager

    def collection_manager(self) -> CollectionManager:
        if not self._collection_manager:
            self._collection_manager = CollectionManager(self)
        return self._collection_manager

    def event_manager(self) -> IEventManager:
        if not self._event_manager:
            self._event_manager = EventManager()
        return self._event_manager

    def task_manager(self) -> ITaskManager:
        if not self._task_manager:
            self._task_manager = TaskManager()
        return self._task_manager

    def close(self):
        if self._collection_manager:
            self._collection_manager.close()



from abc import ABCMeta, abstractmethod

from album.core.api.model.collection_solution import ICollectionSolution
from album.core.api.model.environment import IEnvironment


class IEnvironmentManager:
    """Manages everything around the environment a solution lives in."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def install_environment(
        self, collection_solution: ICollectionSolution
    ) -> IEnvironment:
        raise NotImplementedError

    @abstractmethod
    def set_environment(self, collection_solution: ICollectionSolution) -> IEnvironment:
        """Resolves the environment the active solution runs in.

        Returns the resolve result of the parent of the active solution.

        """
        raise NotImplementedError

    @abstractmethod
    def remove_environment(self, environment: IEnvironment):
        """Removes an environment."""
        raise NotImplementedError

    @abstractmethod
    def run_scripts(self, environment: IEnvironment, scripts, pipe_output=True):
        """Runs the solution in the target environment

        Args:
            scripts:
                List of he scripts calling the solution(s)
            environment:
                The virtual environment used to run the scripts
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.
        """
        raise NotImplementedError

    @abstractmethod
    def get_conda_manager(self):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from typing import Optional, List

from album.core.api.model.database import IDatabase
from album.runner.core.api.model.coordinates import ICoordinates


class ICatalogIndex(IDatabase):
    """Interface handling the index of all solutions in a catalog."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def create(self):
        raise NotImplementedError

    @abstractmethod
    def is_empty(self, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def update_name_version(self, name: str, version: str, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def get_name(self, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def get_version(self, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def get_all_solutions(self, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def get_solution(self, solution_id: int, close: bool = True) -> Optional[dict]:
        raise NotImplementedError

    def get_solution_by_coordinates(
        self, coordinates: ICoordinates, close: bool = True
    ) -> Optional[dict]:
        """Resolves a solution by its name, version and group.

        Args:
            close:
                if specified closes the connection after execution
            coordinates:
                The group affiliation, name, and version of the solution.

        Returns:
            None or row not found.

        """
        raise NotImplementedError

    @abstractmethod
    def get_solution_by_doi(self, doi: str, close: bool = True) -> Optional[dict]:
        """Resolves a solution by its DOI.

        Args:
            close:
                if specified closes the connection after execution
            doi:
                The doi to resolve for.

        Returns:
            None or a node if any found.

        Raises:
            RuntimeError if the DOI was found more than once.
                         if the node found is not a leaf

        """
        raise NotImplementedError

    def get_all_solution_versions(
        self, group: str, name: str, close: bool = True
    ) -> Optional[List[dict]]:
        raise NotImplementedError

    @abstractmethod
    def remove_solution(self, solution_id: int, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def remove_solution_by_group_name_version(
        self, coordinates: ICoordinates, close: bool = True
    ):
        raise NotImplementedError

    @abstractmethod
    def update(
        self, coordinates: ICoordinates, solution_attrs: dict, close: bool = True
    ):
        """Updates a catalog to include a solution as a node with the attributes given.
         Updates exiting nodes if node already present in tree.

        Args:
            close:
                if specified closes the connection after execution
            coordinates:
                The coordinates of the solution.
            solution_attrs:
                The solution attributes. Must hold group, name, version.

        """
        raise NotImplementedError

    @abstractmethod
    def save(self):
        raise NotImplementedError

    @abstractmethod
    def export(self, path, export_format="JSON", close: bool = True):
        """Exports the index tree to disk.

        Args:
            close:
                if specified closes the connection after execution
            path:
                The path to store the export to.
            export_format:
                The format to save to. Choose from ["JSON"]. (Default: JSON)

        Raises:
            NotImplementedError if the format is not supported.

        """
        raise NotImplementedError



from abc import ABCMeta, abstractmethod


class IDeployManager:
    """Class handling the deployment process.

    During deployment, a solution file will be requested to be added to a catalog. This catalog must be configured or
    specified in the solution file.
    When deploying to a remote catalog, deployment happens via a merge request to the git repository of the catalog.
    A deployment can also be requested to a catalog only existing locally.
    In this case, no merge request will be created!

    Notes:
        Git credentials required when deploying to a remote catalog!

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def deploy(
        self,
        deploy_path: str,
        catalog_name: str,
        dry_run: bool,
        push_options=None,
        git_email: str = None,
        git_name: str = None,
        force_deploy: bool = False,
        changelog: str = "",
    ):
        """Function corresponding to the `deploy` subcommand of `album`.

        Generates the yml for a album and creates a merge request to the catalog only
        including the yaml and solution file.

        Args:
            force_deploy:
                Force overwrites a existing solution during deployment. Only for local catalogs.
            deploy_path:
                Path to a directory or a file.
                If directory: Must contain "solution.py" file.
            catalog_name:
                The catalog to deploy to. Either specify via argument in deploy-call, via url in solution or use
                default catalog.
            dry_run:
                When set, prepares deployment in local src of the catlog (creating zip, docker, yml),
                but not adding to the catalog src.
            push_options:
                Push options for the catalog repository.
            git_email:
                The git email to use. (Default: systems git configuration)
            git_name:
                The git user to use. (Default: systems git configuration)
            changelog:
                The change associated with this version of a solution compared to the last version.

        """
        raise NotImplementedError

    @abstractmethod
    def undeploy(
        self,
        solution_to_resolve: str,
        catalog_name: str,
        dry_run: bool,
        push_options=None,
        git_email: str = None,
        git_name: str = None,
    ):
        """Function corresponding to the `deploy` subcommand of `album`.

        Generates the yml for a album and creates a merge request to the catalog only
        including the yaml and solution file.

        Args:
            solution_to_resolve:
                Solution identifier which should be removed (group:name:version).
            catalog_name:
                The catalog to deploy to. Either specify via argument in deploy-call, via url in solution or use
                default catalog.
            dry_run:
                When set, prepares deployment in local src of the catlog (creating zip, docker, yml),
                but not adding to the catalog src.
            push_options:
                Push options for the catalog repository.
            git_email:
                The git email to use. (Default: systems git configuration)
            git_name:
                The git user to use. (Default: systems git configuration)

        """
        raise NotImplementedError



import errno
import json
import os
import platform
import random
import shutil
import stat
import sys
import tempfile
import zipfile
from pathlib import Path
from typing import Optional

import yaml

from album.runner import album_logging

module_logger = album_logging.get_active_logger
enc = sys.getfilesystemencoding()

win_shell = None


def get_dict_from_yml(yml_file):
    """Reads a dictionary from a file in yml format."""
    with open(yml_file, "r") as yml_f:
        d = yaml.safe_load(yml_f)

    if not isinstance(d, dict):
        raise TypeError("Yaml file %s invalid!" % str(yml_file))

    return d


def write_dict_to_yml(yml_file, d):
    """Writes a dictionary to a file in yml format."""
    yml_file = Path(yml_file)
    create_path_recursively(yml_file.parent)

    with open(yml_file, "w+") as yml_f:
        yml_f.write(yaml.dump(d, Dumper=yaml.Dumper))

    return True


def get_dict_entry(d, key, allow_none=True, message=None):
    """Receive an entry from a dictionary.

    Args:
        d:
            The yml dictionary.
        key:
            The key.
        allow_none:
            boolean flag indicating whether to allow key to exist or not.
        message:
            When allow_none false this is the message printed in the exception.


    Returns:
        The value or None

    Raises:
        KeyError if allow_none is False and key not found.

    """
    try:
        val = d[key]
    except KeyError:
        val = None
        if not allow_none:
            raise KeyError(message)

    return val


def write_dict_to_json(json_file, d):
    """Writes dictionary to JSON file."""
    json_file = Path(json_file)
    create_path_recursively(json_file.parent)

    with open(json_file, "w+") as json_f:
        json_f.write(json.dumps(d))

    return True


def get_dict_from_json(json_file):
    """Reads dictionary from JSON file."""
    json_file = Path(json_file)

    with open(json_file, "r") as json_f:
        d = json.load(json_f)

    return d


def folder_empty(path) -> bool:
    """Returns true when a given folder is empty else false."""
    path = Path(path)

    if path.exists():
        if path.is_dir() and os.listdir(path) == []:
            return True
        return False
    return True


def list_files_recursively(path, root=None, relative=False) -> list:
    """Lists all files in a repository recursively"""
    path = Path(path)
    if not root:
        root = path
    files_list = []

    for cur_root, dirs, files in os.walk(path):
        cur_root = Path(cur_root)

        for d in dirs:
            files_list += list_files_recursively(cur_root.joinpath(d), root, relative)
        for fi in files:
            if relative:
                files_list.append(cur_root.joinpath(fi).relative_to(root))
            else:
                files_list.append(cur_root.joinpath(fi))
        break

    return files_list


def create_empty_file_recursively(path_to_file):
    """Creates an empty file. Creates missing parent folders."""
    p = Path(path_to_file)
    p.parent.mkdir(parents=True, exist_ok=True)
    p.touch(exist_ok=True)

    return True


def create_path_recursively(path):
    """Creates a path. Creates missing parent folders."""
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)

    return True


def create_paths_recursively(paths):
    """Create paths. Creates missing parent folders."""
    for p in paths:
        create_path_recursively(p)


def copy_folder(folder_to_copy, destination, copy_root_folder=True, force_copy=False):
    """Copies a folder to a destination.

    Args:
        folder_to_copy:
            The folder to copy
        destination:
            The destination folder to copy to
        copy_root_folder:
            boolean value. if true copies the root folder in the target destination.
            Else all files in the folder to copy.
        force_copy:
            boolean value. If true, removes the destination folder before copying.

    Returns:

    """
    folder_to_copy = Path(folder_to_copy)
    destination = Path(destination)

    if os.path.exists(destination) and os.path.samefile(folder_to_copy, destination):
        return destination

    if copy_root_folder:
        destination = destination.joinpath(folder_to_copy.name)

    if force_copy:
        force_remove(destination)

    create_path_recursively(destination)

    for root, dirs, files in os.walk(folder_to_copy):
        root = Path(root)

        for d in dirs:
            copy_folder(
                root.joinpath(d), destination.joinpath(d), copy_root_folder=False
            )
        for fi in files:
            copy(root.joinpath(fi), destination)
        break

    return destination


def copy(file, path_to) -> Path:
    """Copies a file A to either folder B or file B. Makes sure folder structure for target exists."""
    file = Path(file)
    path_to = Path(path_to)

    if os.path.exists(path_to) and os.path.samefile(file, path_to):
        return path_to

    create_path_recursively(path_to.parent)

    return Path(shutil.copy(file, path_to))


def copy_in_file(file_content, file_path) -> Path:
    """Copies a stream content to a file."""
    file_path = Path(file_path)
    create_path_recursively(file_path.parent)
    if file_path.is_file():
        file_path.unlink()
    with open(file_path, "w") as script_file:
        script_file.write(file_content)
    return file_path


def zip_folder(folder_path, zip_archive_file):
    """Creates a zip archive of a given folder."""
    zip_archive_file = Path(zip_archive_file)

    if zip_archive_file.suffix == ".zip":  # remove suffix as it appends automatically
        zip_archive_file = zip_archive_file.with_suffix("")

    create_path_recursively(zip_archive_file.parent)

    return shutil.make_archive(str(zip_archive_file), "zip", folder_path)


def zip_paths(paths_to_include, zip_archive_file, tmp_dir=None):
    """Creates a zip archive including all given files. Copies the data into a tmp directory first."""

    with tempfile.TemporaryDirectory(dir=tmp_dir) as tmp_folder:
        for file in paths_to_include:
            file = Path(file)
            if file.is_file():
                copy(file, tmp_folder)
            elif file.is_dir():
                copy_folder(file, tmp_folder)
            else:
                raise FileNotFoundError("Could not find file %s" % str(file))

        return zip_folder(tmp_folder, zip_archive_file)


def unzip_archive(zip_archive_file, target_folder=None):
    """Unzips a archive file to a given folder or the folder where it is located (default)."""
    zip_archive_file = Path(zip_archive_file)
    target_folder = Path(target_folder) if target_folder else zip_archive_file.parent

    create_path_recursively(target_folder)

    shutil.unpack_archive(str(zip_archive_file), str(target_folder))

    return target_folder


def force_remove(path, warning=True):
    path = Path(path)
    if path.exists():
        try:
            if path.is_file():
                try:
                    path.unlink()
                except PermissionError:
                    handle_remove_readonly(os.unlink, path, sys.exc_info())
            else:
                shutil.rmtree(
                    str(path), ignore_errors=False, onerror=handle_remove_readonly
                )
        except PermissionError as e:
            module_logger().warn("Cannot delete %s." % str(path))
            if not warning:
                raise e


def remove_link(link_target):
    """Removes a link from the file system."""
    if link_target:
        dispose_op = getattr(link_target, "dispose", None)
        if callable(dispose_op):
            dispose_op()
        else:
            raise RuntimeError("Path doesn't seem to be a link")
        force_remove(link_target)


def handle_remove_readonly(func, path, exc):
    """Changes readonly flag of a given path."""
    excvalue = exc[1]
    if func in (os.rmdir, os.remove, os.unlink) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)
    else:
        raise


def rand_folder_name(f_len=8):
    """Creates random folder name of lenght given."""
    characters = "abcdefghijklmnopqrstuvwxyz0123456789_"
    s = []
    for i in range(0, f_len):
        s.append(characters[random.randint(0, len(characters) - 1)])
    return "".join(s)


def check_zip(path):
    """Checks a given zip file."""
    return zipfile.is_zipfile(path)


def construct_cache_link_target(
    point_to_base: Path, point_from, point_to, create=True
) -> Optional[Path]:
    """Constructs a link from point_from to point_to.

    Args:
        point_to_base:
            Base folder where to point to.
        point_from:
            Path where to point from.
        point_to:
            Folder name in point_to_base where to point to.
        create:

    Returns:

    """
    operation_system = platform.system().lower()
    root = point_to_base.joinpath(point_to)

    if "windows" in operation_system:
        point_from = str(point_from) + ".lnk"
        shortcut = Path(point_from)
        if shortcut.exists():
            return _get_shortcut_target(shortcut)
        if create:
            point_to = root.joinpath(_next_free_pointer_number(root))

            create_path_recursively(point_to)
            create_path_recursively(shortcut.parent)

            _create_shortcut(shortcut, target=point_to)
            resolve = point_to.absolute()
            return resolve
    else:
        if os.path.islink(point_from):
            return Path(point_from).resolve()
        if create:
            point_to = root.joinpath(_next_free_pointer_number(root))

            create_path_recursively(point_to)
            create_path_recursively(Path(point_from).parent)

            # point_from -> point_to
            os.symlink(str(point_to), point_from, target_is_directory=True)
            return point_to.resolve()


def _next_free_pointer_number(root):
    root = Path(root)

    # determine next free point_to number
    i = 0
    while root.joinpath("%s" % i).exists():
        i += 1

    return str(i)


def get_link_target(link: Path):
    operation_system = platform.system().lower()
    if "windows" in operation_system:
        # pylnk3 is windows only
        link = str(link) + ".lnk"
        shortcut = Path(link)
        if shortcut.exists():
            return _get_shortcut_target(shortcut)
    else:
        if Path(link).exists():
            return Path(link).resolve()
    return None


def _create_shortcut(shortcut_path, target):
    shortcut_path = Path(shortcut_path)
    if not shortcut_path.parent.exists():
        shortcut_path.parent.mkdir(parents=True, exist_ok=True)
    sh = _get_global_win_shell()
    wscript = sh.CreateShortCut(str(shortcut_path.absolute()))
    wscript.TargetPath = str(Path(target).absolute())
    wscript.save()


def _get_shortcut_target(shortcut_path):
    shortcut_path = Path(shortcut_path)
    if not shortcut_path.parent.exists():
        raise LookupError("Shortcut %s doesn't exist." % shortcut_path)

    sh = _get_global_win_shell()
    wscript = sh.CreateShortCut(str(shortcut_path.absolute()))
    return Path(wscript.TargetPath)


def _get_global_win_shell():
    global win_shell
    import win32com.client

    if win_shell is None:
        win_shell = win32com.client.Dispatch("Wscript.Shell")
    return win_shell



from album.ci.controller.release_manager import ReleaseManager


def configure_repo(release_manager: ReleaseManager, args):
    release_manager.configure_repo(args.ci_user_name, args.ci_user_email)


def configure_ssh(release_manager: ReleaseManager, args):
    release_manager.configure_ssh(args.ci_project_path)


def zenodo_publish(release_manager: ReleaseManager, args):
    release_manager.zenodo_publish(
        args.branch_name, args.zenodo_base_url, args.zenodo_access_token
    )


def zenodo_upload(release_manager: ReleaseManager, args):
    release_manager.zenodo_upload(
        args.branch_name,
        args.zenodo_base_url,
        args.zenodo_access_token,
        args.report_file,
    )


def update_index(release_manager: ReleaseManager, args):
    release_manager.update_index(args.branch_name, args.doi, args.deposit_id)


def commit_changes(release_manager: ReleaseManager, args):
    release_manager.commit_changes(
        args.branch_name, args.ci_user_name, args.ci_user_email
    )


def merge(release_manager: ReleaseManager, args):
    release_manager.merge(
        args.branch_name,
        args.dry_run,
        args.push_option,
        args.ci_user_name,
        args.ci_user_email,
    )



from album.core.api.controller.event_manager import IEventManager
from album.core.api.model.event import IEvent
from album.runner.core.api.model.coordinates import ICoordinates


class EventManager(IEventManager):
    """Class for handling events."""

    class EventCallback:
        def __init__(self, event_name, method, solution_id):
            self.event_name = event_name
            self.method = method
            self.solution_id = solution_id

    def __init__(self):
        self.event_callbacks = []

    def add_listener(
        self, event_name, callback_method, coordinates: ICoordinates = None
    ):
        res = self._filter_callbacks(
            event_name, coordinates=coordinates, callback_method=callback_method
        )
        if not res:
            self.event_callbacks.append(
                EventManager.EventCallback(event_name, callback_method, coordinates)
            )

    def remove_listener(
        self, event_name, callback_method, coordinates: ICoordinates = None
    ):
        res = self._filter_callbacks(
            event_name, callback_method=callback_method, coordinates=coordinates
        )
        for c in res:
            self.event_callbacks.remove(c)

    def _filter_callbacks(
        self, event_name, coordinates: ICoordinates = None, callback_method=None
    ):
        if coordinates and callback_method:
            return [
                callback
                for callback in self.event_callbacks
                if callback.event_name == event_name
                and callback.method == callback_method
                and (not callback.solution_id or callback.solution_id == coordinates)
            ]
        if coordinates:
            return [
                callback
                for callback in self.event_callbacks
                if callback.event_name == event_name
                and (not callback.solution_id or callback.solution_id == coordinates)
            ]
        if callback_method:
            return [
                callback
                for callback in self.event_callbacks
                if callback.event_name == event_name
                and callback.method == callback_method
            ]
        return [
            callback
            for callback in self.event_callbacks
            if callback.event_name == event_name
        ]

    def publish(self, event: IEvent, coordinates: ICoordinates = None):
        for callback in self._filter_callbacks(event.name(), coordinates=coordinates):
            callback.method(event)



import os
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Generator
from urllib.parse import urlparse

import git
from git import Repo

from album.core.utils.operations.file_operations import (
    force_remove,
    create_path_recursively,
    folder_empty,
)
from album.core.utils.operations.url_operations import is_url
from album.runner import album_logging

module_logger = album_logging.get_active_logger


def checkout_branch(git_repo, branch_name):
    """Checks out a branch on a repository.

    First, local refs are taken, then refs pointing to origin.

    Args:
        git_repo:
            The repository
        branch_name:
            The name of the branch to check out.

    Returns:
        The head of the branch

    Raises:
        IndexError when the branch is no where available.

    """
    try:  # checkout branch locally
        module_logger().debug(
            "Found the following branches locally: \n %s..."
            % "\n".join([h.name for h in git_repo.heads])
        )
        head = git_repo.heads[branch_name]
        head.checkout()
        try:
            head.repo.git.pull()
        except git.GitCommandError:
            # only for request based catalogs this is important. Not failing here should be safe.
            module_logger().warning("Cannot pull from branch. Assuming up to date!")
        return head
    except IndexError as e:
        module_logger().debug("Branch name not in local repository! Checking origin...")
        for remote in git_repo.remotes:  # checkout branch remote and track
            try:
                module_logger().debug("Trying remote: %s..." % remote.name)
                head = remote.refs[branch_name].checkout("--track")
                return head
            except IndexError:
                module_logger().debug("Not found! Trying next...")
        module_logger().debug("Nothing left to try...")
        raise IndexError('Branch "%s" not in repository!' % branch_name) from e


def retrieve_files_from_head(head, pattern, number_of_files=1, relative=False):
    """Extracts a file with a pattern given a branch (or head) of a repository.

    Args:
        number_of_files:
            The number of files to expect to be found when using the given pattern. Use -1 for arbitrary many.
        head:
            The branch (or more general head) of the catalog repository where the solution file was committed in.
        pattern:
            The pattern the path of the file to retrieve.
        relative:
            Whether the returned paths are relative or absolute.

    Returns:
        The absolute path of the solution directory (in the catalog repository).

    """
    pattern = str(pattern)

    pattern_matches = head.repo.git.ls_files(pattern)

    res = []
    for file in pattern_matches.split("\n"):
        # unfortunately git on windows internally uses linux-separator as path separator.
        # Paths therefore might have the wrong separator.
        file = file.replace("/", os.path.sep)
        module_logger().debug("Found file: %s..." % file)
        if relative:
            res.append(file)
        else:
            res.append(os.path.join(head.repo.working_tree_dir, file))

    if not res:
        raise RuntimeError(
            'Head "%s" does not hold pattern "%s"! Aborting...' % (head.name, pattern)
        )

    if number_of_files > 0:
        if len(res) != number_of_files:
            raise RuntimeError(
                'Head "%s" holds pattern "%s" %s times, but expected %s. Aborting...'
                % (head.name, pattern, len(res), number_of_files)
            )

    return res


def retrieve_files_from_head_last_commit(head, pattern, option="", number_of_files=1):
    """Extracts a file with a "startswith" pattern given a branch (or head) of a repository, only factoring in
    files that were changed in the last commit.

    Args:
        number_of_files:
            The number of files to expect to be found when using the given pattern. Use -1 for arbitrary many.
        option:
            Specify "startswith" to do a prefix comparison instead of a regular expression matching
        head:
            The branch (or more general head) of the catalog repository where the solution file was committed in.
        pattern:
            The pattern the path of the file to retrieve should start with.

    Returns:
        The absolute path of the solution file (in the catalog repository).

    Raises:
        RuntimeError when
            a) there is no parent commit in the head
            b) No solution file is included in the latest commit of the given head
            c) There are more than a single solution file committed in the latest commit.

    """
    pattern = str(pattern)
    parent = head.commit.parents[0] if head.commit.parents else None
    module_logger().debug("Found head commit: %s..." % parent)
    module_logger().debug("Summary message: %s..." % parent.summary)

    if not parent:
        raise RuntimeError("Cannot execute diff since there is only a single commit!")

    diff = head.commit.diff(parent)

    abs_path_solution_file = []
    for file in diff:
        # unfortunately git on windows internally uses linux-separator as path separator.
        # Paths therefore might have the wrong separator.
        path = file.b_path.replace("/", os.path.sep)
        module_logger().debug("Found file in commit diff to parent: %s..." % path)
        module_logger().debug("Matching pattern %s..." % pattern)

        if option == "startswith":
            if path.startswith(pattern):
                module_logger().debug(
                    "Found file matching pattern %s: %s..." % (pattern, path)
                )
                abs_path_solution_file.append(
                    os.path.join(head.repo.working_tree_dir, path)
                )
        else:
            if re.search(pattern, path):
                module_logger().debug(
                    "Found file matching pattern %s: %s..." % (pattern, path)
                )
                abs_path_solution_file.append(
                    os.path.join(head.repo.working_tree_dir, path)
                )

    if not abs_path_solution_file:
        raise RuntimeError(
            'Head "%s" does not hold pattern "%s"! Aborting...' % (head.name, pattern)
        )

    if number_of_files > 0:
        if len(abs_path_solution_file) != number_of_files:
            raise RuntimeError(
                'Head "%s" holds pattern "%s" %s times, but expected %s. Aborting...'
                % (head.name, pattern, len(abs_path_solution_file), number_of_files)
            )

    return abs_path_solution_file


def _add_files(repo, file_paths) -> bool:
    """Add files to the repo in the branch currently checked out."""
    if repo.index.diff(None) or repo.untracked_files:
        module_logger().info("Preparing committing...")
        for file_path in file_paths:
            module_logger().debug("Adding file %s..." % file_path)
            # todo: nice catching here?
            repo.git.add(file_path)
        return True
    return False


def add_files_commit_and_push(
    head,
    file_paths,
    commit_message,
    push=False,
    email=None,
    username=None,
    push_option_list=None,
    force=False,
):
    """Adds files in a given path to a git head and commits.

    Args:
        push_option_list:
            options used for pushing. See https://docs.gitlab.com/ee/user/project/push_options.html. Expects a list.
        head:
            The head of the repository
        file_paths:
            The path of the files to add
        commit_message:
            The commit message
        push:
            Boolean option to switch on/off pushing to repository remote
        username:
            The git user to use. (Default: systems git configuration)
        email:
            The git email to use. (Default: systems git configuration)
        force:
            whether to use force push or not

    Raises:
        RuntimeError when no files are in the index

    """
    if push_option_list is None or push_option_list == []:
        push_options = []
    else:
        push_options = ["-o %s" % o for o in push_option_list]

    repo = head.repo

    if email or username:
        configure_git(repo, email, username)

    if _add_files(repo, file_paths):

        # build command
        cmd_option = ["--set-upstream"]
        cmd = cmd_option + push_options
        if force:
            cmd = cmd + ["-f"]

        remote_name = get_remote_name(repo)

        cmd = cmd + [remote_name, head]

        module_logger().debug(
            "Running command: repo.git.push(%s)..." % (", ".join(str(x) for x in cmd))
        )

        # commit
        repo.git.commit(m=commit_message)

        if push:
            module_logger().info("Preparing pushing...")
            repo.git.push(cmd)

    else:
        raise RuntimeError("Diff shows no changes to the repository. Aborting...")


def add_tag(repo, tag):
    """Adds a tag to the most recent commit.

    Args:
        repo:
            The repository
        tag:
            The tag associated with the commit
    """
    repo.git.tag("-a", tag, "-f", "-m", "")
    remote_name = get_remote_name(repo)
    repo.git.push([remote_name, tag, "-f"])


def get_remote_name(repo):
    try:
        remote_name = repo.remote().refs.HEAD.remote_name
    except AttributeError:
        remote_name = "origin"
    return remote_name


def remove_files(head, file_paths):
    """Adds files in a given path to a git head and commits.

    Args:
        head:
            The head of the repository
        file_paths:
            The path of the files to remove
    """
    repo = head.repo

    if repo.index.diff(None) or repo.untracked_files:
        for file_path in file_paths:
            module_logger().debug("Removing file %s..." % file_path)
            repo.git.rm([file_path, "-r", "--ignore-unmatch", "--cached"])


def remove_tag(repo, tag):
    """Adds files in a given path to a git head and commits.

    Args:
        repo:
            The repository
        tag:
            The tag to remove
    """
    remote_name = get_remote_name(repo)
    repo.git.tag("-d", tag)
    repo.git.push("--delete", remote_name, tag)


def get_tags(repo):
    tags = [
        tag.name for tag in sorted(repo.tags, key=lambda t: t.commit.committed_datetime)
    ]
    tags.reverse()
    return tags


def revert(repo, tag, files: list):
    for file in files:
        repo.git.reset(tag, "--", file)
        repo.git.checkout("--", file)
        repo.git.clean("-fd", file)


def configure_git(repo, email, username):
    """Configures email and username to use for git operations for the given repo.

    Args:
        repo:
            The repo to configure.
        email:
            The email to use.
        username:
            The username to use.

    Returns:
        repo:
            The configured repository.

    """
    if username:
        module_logger().info('Set username to "%s"' % username)
        repo.config_writer().set_value("user", "name", username).release()
    if email:
        module_logger().info('Set email to "%s"' % email)
        repo.config_writer().set_value("user", "email", email).release()
    if not username and not email:
        module_logger().warning("Neither username nor email given! Doing nothing...")

    return repo


# todo: write test
def create_new_head(repo, name):
    """Force creates a new head of a given name in a repository and returns the head."""

    if name in repo.heads:
        git.Head.delete(repo, name, force=True)
    new_head = repo.create_head(name)

    return new_head


def download_repository(
    repo_url, git_folder_path, force_download=True, update=True
) -> Repo:
    """Downloads or updates the repository behind a url, returns repository object on success.

    If repository already cached, head is detached to origin HEAD for a clean start for new branches.

    Args:
        repo_url:
            The URL to the git.
        git_folder_path:
            The complete path to clone to
        force_download:
            Boolean, indicates whether to force delete existing folder before cloning
        update:
            Flag to indicate whether to hard reset the repo upon initialization or not.

    Returns:
        The repository object

    """
    git_folder_path = Path(git_folder_path)
    Path.mkdir(git_folder_path, parents=True, exist_ok=True)

    # update existing repo or clone new repo
    if Path.exists(git_folder_path.joinpath(".git")):
        module_logger().info("Found existing repository in %s..." % git_folder_path)

        if update:
            repo = init_repository(git_folder_path)
        else:
            repo = git.Repo(git_folder_path)
    else:

        if force_download:
            force_remove(git_folder_path)

        module_logger().info(
            "Download repository from %s in %s..." % (repo_url, git_folder_path)
        )
        repo = git.Repo.clone_from(repo_url, git_folder_path)

    return repo


def checkout_files(repo, files_to_download):
    for file in files_to_download:
        repo.git.restore(file, staged=True)
        repo.git.checkout(file)


@contextmanager
def clone_repository_sparse(
    repo_url, branch_name, target_repo_path
) -> Generator[Repo, None, None]:
    """Clones a repository branch to a given path.

    Args:
        repo_url:
            The url to the repository.
        branch_name:
            The branch name.
        target_repo_path:
            The target path on the disk.

    """
    git_folder_path = Path(target_repo_path)
    force_remove(git_folder_path)
    create_path_recursively(git_folder_path)

    module_logger().debug(
        "Cloning repository without history from %s into %s..."
        % (repo_url, git_folder_path)
    )
    repo = git.Repo.clone_from(
        repo_url,
        git_folder_path,
        branch=branch_name,
        no_checkout=True,
        depth=1,
        no_tags=True,
        single_branch=True,
    )
    yield repo
    repo.close()


@contextmanager
def clone_repository(src, target_repo_path, force=False) -> Generator[Repo, None, None]:
    """Clones the full repository

    Args:
        src:
            The url to the repository.
        target_repo_path:
            The target path on the disk.
        force:
            boolean value. If true deletes target folder first


    """
    if not folder_empty(target_repo_path):
        if force:
            force_remove(target_repo_path)
        else:
            raise RuntimeError('Target folder "%s" not empty!' % str(target_repo_path))

    create_path_recursively(target_repo_path)
    repo = git.Repo.clone_from(src, target_repo_path)

    yield repo
    repo.close()


def init_repository(path):
    """Initializes a repository to the origin reference. Thereby discarding all changes made to the repository.

    Usually this means checking out the origin "main" branch, but this depends on the repository configuration.

    Args:
        path:
            The path to the repository to init/reset.

    Returns:
        the repository object.

    """
    module_logger().info("Initialize repository...")
    path = Path(path)

    repo = git.Repo(path)

    # update the remote to get latest changes on all remotes (pushes, HEAD pointer change, reverts, etc.)
    repo.remote().update()

    # remove all eventual changes made local
    remote_head = get_local_remote_ref_head(repo)
    checkout_main(repo, remote_head.name)
    clean_repository(repo, remote_head.name)

    # checkout remote HEAD for a clean start for new branches
    remote_head.checkout()

    return repo


def create_bare_repository(target):
    create_path_recursively(target)

    repo = git.Repo.init(target, bare=True)

    # ref HEAD to main
    repo.git.symbolic_ref("HEAD", "refs/heads/main")

    repo.close()


def create_repository(target):
    create_path_recursively(target)

    repo = git.Repo.init(target)

    # ref HEAD to main
    repo.git.symbolic_ref("HEAD", "refs/heads/main")

    return repo


def clean_repository(repo, target_head_name=None):
    """Resets all changes made in the current repository"""
    if not target_head_name:
        target_head_name = get_local_remote_ref_head(repo).name
    remote_name = repo.remote().name
    remote_ref = remote_name + "/" + target_head_name
    # reset to current remote HEAD reference name
    repo.git.reset(["--hard", remote_ref])
    # remove all leftover-untracked files (if any)
    repo.git.clean("-fd")


def get_local_remote_ref_head(repo):
    if repo.remote().refs:
        try:
            remote_head = repo.git.remote(["set-head", "origin", "-a"])
            remote_main_name = remote_head.split(" ")[-1]
        except git.GitCommandError:
            remote_main_name = "main"
        head = repo.heads[remote_main_name]
    else:
        head = repo.head.ref
    return head


def checkout_main(repo, main_name=None):
    """Checks out the main branch of the repository locally. Note: must not be called "main"!"""
    if not main_name:
        head = get_local_remote_ref_head(repo)
    else:
        head = repo.heads[main_name]

    try:
        head.checkout()
    except git.GitCommandError:
        pass

    return head


def retrieve_default_mr_push_options(repo_url) -> list:
    """Returns the default push option for the given repository host if available (e.g. gitlab, github)

    Args:
        repo_url:
            The source url of the catalog. Netloc determines default push option.

    Returns:
        The default push option to directly create a merge request when pushed to origin.

    """
    if is_url(str(repo_url)):
        parsed_url = urlparse(repo_url)

        if parsed_url.netloc.startswith("gitlab"):
            return ["merge_request.create"]
        else:
            return []
    else:
        return []



import argparse
import sys
import traceback

import pkg_resources

from album import core
from album.api import Album
from album.commandline import (
    add_catalog,
    remove_catalog,
    deploy,
    install,
    repl,
    run,
    search,
    test,
    update,
    clone,
    upgrade,
    index,
    uninstall,
    info,
    undeploy,
)
from album.core.utils.subcommand import SubProcessError
from album.runner.album_logging import (
    debug_settings,
    get_active_logger,
    LogLevel,
    to_loglevel,
)


def main():
    """Entry points of `album`."""

    parser = create_parser()

    get_active_logger().debug("Parsing base album call arguments...")
    namespace, args = parser.parse_known_args()
    __handle_args(namespace, args, parser)


def __handle_args(namespace, args, parser):
    """Handles all arguments provided after the album command."""
    level = namespace.log
    print_json = getattr(namespace, "json", False)
    __run_subcommand(namespace, args, parser, level, print_json)


def _capture_output():
    logger = get_active_logger()
    logger.handlers.clear()


def _handle_exception(e, silent: bool = False):
    if not silent:
        get_active_logger().error("album command failed: %s" % str(e))
        get_active_logger().debug(traceback.format_exc())
    if type(e) is int:
        sys.exit(e)
    elif hasattr(e, "exit_status"):
        sys.exit(e.exit_status)
    else:
        sys.exit(e)


def __run_subcommand(namespace, args, parser, level: LogLevel, print_json):
    """Calls a specific album subcommand."""
    album_command = ""
    try:
        album_command = sys.argv[1]  # album command always expected at second position
    except IndexError:
        parser.error("Please provide a valid action!")
    get_active_logger().debug("Running %s subcommand..." % album_command)
    sys.argv = [sys.argv[0]] + args

    # Makes sure album is initialized.
    album_instance = create_album_instance(level)
    if print_json:
        _capture_output()
    get_active_logger().info("album version %s" % core.__version__)

    try:
        if hasattr(namespace, "func"):
            album_instance.load_or_create_collection()
            namespace.func(album_instance, namespace)  # execute entry point function
        else:
            get_active_logger().error("Invalid argument(s): %s" % args)
            sys.exit(1)
    except KeyboardInterrupt:
        get_active_logger().error("Album command canceled.")
        sys.exit(1)
    except SubProcessError as e:
        _handle_exception(e, silent=True)
    except Exception as e:
        _handle_exception(e)


def create_album_instance(level: LogLevel) -> Album:
    return Album.Builder().log_level(level).build()


def create_parser():
    """Creates a parser for all known album arguments."""
    parser = AlbumParser()
    parser_creators = []
    for entry_point in pkg_resources.iter_entry_points("console_parsers_album"):
        try:
            parser_creators.append(entry_point.load())
        except Exception as e:
            get_active_logger().error(
                "Cannot load console parser %s" % entry_point.name
            )
            get_active_logger().debug(str(e))
    for parse_creator in parser_creators:
        parse_creator(parser)
    return parser.parser


def create_test_parser(parser):
    parser.create_file_command_parser(
        "test", test, "execute the test routine of a solution."
    )


def create_index_parser(parser):
    parser.create_command_parser(
        "index", index, "print the index of the local album collection."
    )


def create_clone_parser(parser):
    p = parser.create_command_parser(
        "clone", clone, "clone an album solution or catalog template."
    )
    p.add_argument(
        "src",
        type=str,
        help="path for the solution file, group:name:version or name of the catalog template.",
    )
    p.add_argument(
        "target_dir",
        type=str,
        help="The target directory where the solution or catalog will be added to. For a catalog, this can also be an empty GIT repository URL.",
    )
    p.add_argument(
        "name", type=str, help="The new name of the cloned solution or catalog."
    )


def create_upgrade_parser(parser):
    p = parser.create_command_parser(
        "upgrade",
        upgrade,
        "upgrade the local collection from the catalog index files. Either all catalogs configured, or a specific one.",
    )
    p.add_argument("catalog", type=str, help="name of the catalog", nargs="?")
    p.add_argument(
        "--dry-run",
        required=False,
        help="Parameter to indicate a dry run and only show what would happen.",
        action="store_true",
    )
    p.add_argument(
        "--override",
        required=False,
        help="renews all installed solutions without re-installation. "
        "Might produce dependency problems as environments are not updated. Handle with care.",
        action="store_true",
    )


def create_update_parser(parser):
    p = parser.create_command_parser(
        "update",
        update,
        "Update the catalog index files. Either all catalogs configured, or a specific one.",
    )
    p.add_argument("catalog", type=str, help="name of the catalog", nargs="?")


def create_remove_catalog_parser(parser):
    p = parser.create_command_parser(
        "remove-catalog",
        remove_catalog,
        "remove a catalog from your local album configuration file.",
    )
    p.add_argument("name", type=str, help="name of the catalog")


def create_add_catalog_parser(parser):
    p = parser.create_command_parser(
        "add-catalog",
        add_catalog,
        "add a catalog to your local album configuration file.",
    )
    p.add_argument("src", type=str, help="src of the catalog")


def create_info_parser(parser):
    parser.create_file_command_parser(
        "info", info, "print information about an album solution."
    )


def create_uninstall_parser(parser):
    p = parser.create_file_command_parser(
        "uninstall", uninstall, "uninstall an album solution."
    )
    p.add_argument(
        "--uninstall-deps",
        required=False,
        help="Boolean to additionally remove all album dependencies. Choose between %s."
        % ", ".join([str(True), str(False)]),
        default=False,
        action="store_true",
    )


def create_install_parser(parser):
    parser.create_file_command_parser("install", install, "install an album solution.")


def create_deploy_parser(parser):
    p = parser.create_file_command_parser("deploy", deploy, "deploy an album solution.")
    p.add_argument(
        "catalog",
        type=str,
        help="Specify a catalog name to deploy to. Must be configured!",
    )
    p.add_argument(
        "--dry-run",
        required=False,
        help="Parameter to indicate a dry run and only show what would happen.",
        action="store_true",
    )
    p.add_argument(
        "--push-option",
        required=False,
        help="Push options for the catalog repository.",
        default=None,
        nargs="+",
    )
    p.add_argument(
        "--git-email",
        required=False,
        help="Email to use for all git operations. If none given, system is required to be proper configured!",
        default=None,
    )
    p.add_argument(
        "--git-name",
        required=False,
        help="Name to use for all git operations. If none given, system is required to be proper configured!",
        default=None,
    )
    p.add_argument(
        "--force-deploy",
        required=False,
        help="When specified,force deploys a solution to a catalog."
        " Useful if the solution has already been deployed once.",
        action="store_false",
    )
    p.add_argument(
        "--changelog",
        required=False,
        help="Description of changes from the previous version of the solution to this version.",
        default=None,
    )


def create_undeploy_parser(parser):
    p = parser.create_file_command_parser(
        "undeploy", undeploy, "undeploy an album solution."
    )
    p.add_argument(
        "catalog",
        type=str,
        help="Specify a catalog name to remove the solution from. Must be configured!",
    )
    p.add_argument(
        "--dry-run",
        required=False,
        help="Parameter to indicate a dry run and only show what would happen.",
        action="store_true",
    )
    p.add_argument(
        "--push-option",
        required=False,
        help="Push options for the catalog repository.",
        default=None,
        nargs="+",
    )
    p.add_argument(
        "--git-email",
        required=False,
        help="Email to use for all git operations. If none given, system is required to be proper configured!",
        default=None,
    )
    p.add_argument(
        "--git-name",
        required=False,
        help="Name to use for all git operations. If none given, system is required to be proper configured!",
        default=None,
    )


def create_repl_parser(parser):
    parser.create_file_command_parser(
        "repl", repl, "get an interactive repl for an album solution."
    )


def create_run_parser(parser):
    parser.create_file_command_parser("run", run, "run an album solution.")


def create_search_parser(parser):
    p = parser.create_command_parser(
        "search", search, "search for an album solution using keywords."
    )
    p.add_argument("keywords", type=str, nargs="+", help="Search keywords")


class ArgumentParser(argparse.ArgumentParser):
    """Override default error method of all parsers to show help of subcommand"""

    def error(self, message):
        self.print_help()
        self.exit(2, "%s: error: %s\n" % (self.prog, message))


class AlbumParser(ArgumentParser):
    def __init__(self):
        super().__init__()
        self.parent_parser = self.create_parent_parser()
        self.parser = self.create_parser()
        self.subparsers = self.parser.add_subparsers(
            title="actions", help="sub-command help"
        )

    @staticmethod
    def create_parent_parser():
        """Parent parser for all subparsers to have the same set of arguments."""
        parent_parser = ArgumentParser(add_help=False)
        # parse logging
        parent_parser.add_argument(
            "--log",
            required=False,
            help="Logging level for your album command. Choose between %s"
            % ", ".join([loglevel.name for loglevel in LogLevel]),
            default=LogLevel(debug_settings()),
            type=(lambda choice: to_loglevel(choice)),
        )
        parent_parser.add_argument(
            "--json",
            required=False,
            help="Adding this parameter prevents the log from being printed to the console."
            " Instead, the result of the command - if present - is printed as JSON.",
            action="store_true",
        )
        parent_parser.add_argument(
            "--version", "-V", action="version", version="%s " % core.__version__
        )
        return parent_parser

    def create_parser(self):
        """Creates the main parser for the album framework."""
        parser = ArgumentParser(
            add_help=True,
            description="album for running, building, and deploying computational solutions",
            parents=[self.parent_parser],
        )
        return parser

    def create_command_parser(self, command_name, command_function, command_help):
        """Creates a parser for a album command, specified by a name, a function and a help description."""
        parser = self.subparsers.add_parser(
            command_name, help=command_help, parents=[self.parent_parser]
        )
        parser.set_defaults(func=command_function)
        return parser

    def create_file_command_parser(self, command_name, command_function, command_help):
        """Creates a parser for a album command dealing with an album file.

        Parser is specified by a name, a function and a help description.
        """
        parser = self.create_command_parser(
            command_name, command_function, command_help
        )
        parser.add_argument("path", type=str, help="path for the solution file")
        return parser


if __name__ == "__main__":
    main()



from abc import ABCMeta, abstractmethod
from queue import Queue

from album.core.api.model.collection_solution import ICollectionSolution
from album.runner.core.api.model.script_creator import IScriptCreator


class IScriptManager:
    __metaclass__ = ABCMeta

    @abstractmethod
    def run_queue(self, queue: Queue):
        """Runs the que. Queue consists of "solution_object and their scripts"-entries. Order matters!

        Args:
            queue:
                The queue object holding entries of solution_object and their scripts

        """
        raise NotImplementedError

    @abstractmethod
    def build_queue(
        self,
        solution: ICollectionSolution,
        queue,
        script_creator: IScriptCreator,
        run_immediately=False,
        argv=None,
    ):
        """Builds the queue of an active-album object.

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            solution:
                The resolve_result of a solution object to build the run-que for.
            queue:
                The que object.
            run_immediately:
                Boolean. When true, a collection is run immediately without solving for further steps and pushing them
                to the queue. Can result in resolving problems in further downstream collections. Necessary for
                branching decisions.
            argv:
                The argument vector being passed to the solution.

        """
        raise NotImplementedError

    @abstractmethod
    def run_solution_script(
        self, resolve_result: ICollectionSolution, script: IScriptCreator
    ):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from enum import unique, IntEnum
from logging import Handler, LogRecord


class ILogHandler(Handler):
    __metaclass__ = ABCMeta

    @abstractmethod
    def emit(self, record: LogRecord) -> None:
        raise NotImplementedError

    @abstractmethod
    def records(self):
        raise NotImplementedError


class ITask:
    __metaclass__ = ABCMeta

    @unique
    class Status(IntEnum):
        WAITING = 0
        RUNNING = 1
        FINISHED = 2
        FAILED = 3

    @abstractmethod
    def id(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def method(self):
        raise NotImplementedError

    @abstractmethod
    def args(self):
        raise NotImplementedError

    @abstractmethod
    def log_handler(self):
        raise NotImplementedError

    @abstractmethod
    def status(self) -> Status:
        raise NotImplementedError

    @abstractmethod
    def set_status(self, status):
        raise NotImplementedError

    @abstractmethod
    def set_log_handler(self, handler):
        raise NotImplementedError

    @abstractmethod
    def set_id(self, task_id: str):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from sqlite3 import Connection, Cursor


class IDatabase:
    __metaclass__ = ABCMeta

    @abstractmethod
    def close_current_connection(self, commit=True):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError

    @abstractmethod
    def get_connection(self) -> Connection:
        raise NotImplementedError

    @abstractmethod
    def get_cursor(self) -> Cursor:
        raise NotImplementedError

    @abstractmethod
    def next_id(self, table_name, close=False) -> int:
        raise NotImplementedError

    @abstractmethod
    def is_created(self, close: bool = True) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_table_empty(self, table, close: bool = True) -> bool:
        raise NotImplementedError

    @abstractmethod
    def create(self):
        raise NotImplementedError

    @abstractmethod
    def is_empty(self, close: bool = True) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_path(self):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod


class IInstallManager:
    """Interface handling the installation and uninstallation process of a solution."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def clean_unfinished_installations(self):
        """Remove traces of any installation that was started but not finished."""
        raise NotImplementedError

    @abstractmethod
    def install(self, solution_to_resolve: str, argv=None):
        """Function corresponding to the `install` subcommand of `album`."""
        raise NotImplementedError

    @abstractmethod
    def uninstall(self, solution_to_resolve: str, rm_dep=False, argv=None):
        """Removes a solution from the disk. Thereby uninstalling its environment and deleting all its downloads.

        Args:
            argv:
                Arguments which should be appended to the script call
            resolve_solution:
                The path, DOI or group-name-version information of the solution to remove.
            rm_dep:
                Boolean to indicate whether to remove parents too.

        """
        raise NotImplementedError



import json
import os
import sys
import tempfile
from pathlib import Path

import yaml
from packaging import version

from album.core.api.model.configuration import IConfiguration
from album.core.api.model.environment import IEnvironment
from album.core.model.default_values import DefaultValues
from album.core.model.link import Link
from album.core.utils import subcommand
from album.core.utils.operations.file_operations import (
    construct_cache_link_target,
    remove_link,
)
from album.core.utils.subcommand import SubProcessError
from album.runner import album_logging
from album.runner.album_logging import debug_settings

module_logger = album_logging.get_active_logger


class CondaManager:
    """Class for handling conda environments.

    The conda class manages the environments a solution is supposed to run in. It provides all features necessary for
    environment creation, deletion, dependency installation, etc.

    Notes:
        An installed \"conda\" program must be available and callable via commandline or powershell.

    """

    def __init__(self, configuration: IConfiguration):
        self._configuration = configuration
        self._conda_executable = self._configuration.conda_executable()

    def _get_install_environment_executable(self):
        return self._conda_executable

    def get_environment_list(self):
        """Returns the available album conda environments."""
        if Path(self._get_base_environment_target_path()).exists():
            return sorted(
                self._get_immediate_subdirectories(
                    self._get_base_environment_target_path()
                )
            )
        else:
            return []

    @staticmethod
    def _get_immediate_subdirectories(a_dir: Path):
        return [
            a_dir.joinpath(name).resolve()
            for name in os.listdir(str(a_dir))
            if os.path.isdir(os.path.join(str(a_dir), name))
        ]

    def _get_base_environment_target_path(self):
        """Gets the first of the paths the conda installation uses to manage its environments."""
        return Path(self._configuration.lnk_path()).joinpath(
            DefaultValues.lnk_env_prefix.value
        )

    def environment_exists(self, environment_name):
        """Checks whether an environment already exists or not.

        Args:
            environment_name:
                The name of the environment in which a solution will run.

        Returns:
            True when environment exists else false.
        """
        environment_list = self.get_environment_list()
        path_expected = self._environment_name_to_path(environment_name, create=False)

        return (
            True
            if (
                path_expected
                and path_expected.resolve() in environment_list
                and os.listdir(path_expected)
            )
            else False
        )

    def _environment_name_to_path(self, environment_name, create=True):
        path = os.path.normpath(
            self._configuration.environments_path().joinpath(environment_name)
        )
        target = construct_cache_link_target(
            self._configuration.lnk_path(),
            point_from=path,
            point_to=DefaultValues.lnk_env_prefix.value,
            create=create,
        )
        if target:
            return Link(target).set_link(path)
        else:
            return None

    def get_environment_path(self, environment_name: str, create: bool = True) -> Link:
        """Gets the environment path for a given environment

        Args:
            environment_name:
                The environment to get the path for.
            create:



        Returns:
            None or the path
        """
        environment_list = self.get_environment_list()
        path_expected = self._environment_name_to_path(environment_name, create)

        if path_expected.resolve() in environment_list:
            return path_expected
        raise LookupError("Could not find environment %s." % environment_name)

    def get_active_environment_name(self):
        """Returns the environment from the active album."""
        conda_list = self.get_info()
        return conda_list["active_prefix_name"]

    def get_active_environment_path(self):
        """Returns the environment form the active album."""
        conda_list = self.get_info()
        path = conda_list["active_prefix"]
        link = construct_cache_link_target(
            self._configuration.lnk_path(),
            point_from=path,
            point_to=DefaultValues.lnk_env_prefix.value,
            create=False,
        )
        if link:
            return link
        else:
            return Link(path)

    def remove_environment(self, environment_name) -> bool:
        """Removes an environment given its name. Does nothing when environment does not exist.

        Args:
            environment_name:
                The name of the environment to remove

        Returns:
            True, when removal succeeded, else False

        """
        if self.get_active_environment_name() == environment_name:
            module_logger().warning("Cannot remove active environment! Skipping...")
            return False

        if not self.environment_exists(environment_name):
            module_logger().warning("Environment does not exist! Skipping...")
            return False

        path = self.get_environment_path(environment_name, create=False)

        try:
            subprocess_args = [
                self._get_install_environment_executable(),
                "env",
                "remove",
                "-y",
                "-q",
                "-p",
                os.path.normpath(path),
            ]

            subcommand.run(subprocess_args, log_output=False)
        except SubProcessError:
            module_logger().debug(
                "Can't delete environment via command line call, deleting the folder next..."
            )
        # try to remove file content if any but don't fail:
        remove_link(path)
        return True

    def get_info(self):
        """Get the info of the conda installation on the corresponding system.

        Returns:
            dictionary corresponding to conda info.
        """
        subprocess_args = [self._get_install_environment_executable(), "info", "--json"]
        output = subcommand.check_output(subprocess_args)
        return json.loads(output)

    def list_environment(self, environment_path):
        """Lists all available conda installation in the given environment.

        Args:
            environment_path:
                The prefix of the environment to list.

        Returns:
            dictionary containing the available packages in the given conda environment.
        """
        subprocess_args = [
            self._get_install_environment_executable(),
            "list",
            "--json",
            "--prefix",
            str(environment_path),
        ]
        output = subcommand.check_output(subprocess_args)
        return json.loads(output)

    def create_environment_from_file(
        self, yaml_path, environment_name, album_api_version=None
    ):
        """Creates a conda environment given a path to a yaml file and its name.

        Args:
            yaml_path:
                The path to the file.
            environment_name:
                The name of the environment. Must be the same as specified in the yml file.
            album_api_version:
                The api version (runner version) in the environment.

        Raises:
            NameError:
                When the file has the wrong format according to its extension.
            ValueError:
                When the file is unreadable or empty.
            RuntimeError:
                When the environment could not be created due to whatever reasons.

        """
        if self.environment_exists(environment_name):
            self.remove_environment(environment_name)

        if not (str(yaml_path).endswith(".yml") or str(yaml_path).endswith(".yaml")):
            raise NameError("File needs to be a yml or yaml file!")

        yaml_path = Path(yaml_path)

        if not (yaml_path.is_file() and yaml_path.stat().st_size > 0):
            raise ValueError("File not a valid yml file!")

        with open(yaml_path, "r") as f:
            content = yaml.safe_load(f)

        self._append_runner_and_install(album_api_version, environment_name, content)

    @staticmethod
    def _append_framework_to_yml(content, album_api_version):
        if version.parse(album_api_version) >= version.parse(
            DefaultValues.first_runner_conda_version.value
        ):
            return CondaManager._append_framework_via_conda_to_yml(
                content, album_api_version
            )
        else:
            return CondaManager._append_framework_via_pip_to_yml(
                content, album_api_version
            )

    @staticmethod
    def _append_framework_via_conda_to_yml(content, album_api_version):
        framework = "conda-forge::%s=%s" % (
            DefaultValues.runner_api_packet_name.value,
            album_api_version,
        )
        if not "dependencies" in content or not content["dependencies"]:
            content["dependencies"] = []
        content["dependencies"].append(framework)
        return content

    @staticmethod
    def _append_framework_via_pip_to_yml(content, album_api_version):
        # There might be environments without pip, so the the dependency content needs
        # to be adjusted to make sure pip is available.
        framework = "%s==%s" % (
            DefaultValues.runner_api_packet_name.value,
            album_api_version,
        )
        if not "dependencies" in content or not content["dependencies"]:
            content["dependencies"] = []
        pip_dict = None
        found_pip_dep = False
        for dep in content["dependencies"]:
            if isinstance(dep, dict):
                if "pip" in dep:
                    pip_dict = dep
            if dep == "pip" or str(dep).startswith("pip="):
                found_pip_dep = True
        if not found_pip_dep:
            content["dependencies"].append(DefaultValues.runner_pip_version.value)
        if not pip_dict:
            pip_dict = {"pip": []}
            content["dependencies"].append(pip_dict)
        pip_dict["pip"].append(framework)
        return content

    def create_environment(self, environment_name, album_api_version=None, force=False):
        """Creates a conda environment with python (latest version) installed.

        Args:
            environment_name:
                The desired environment name.
            force:
                If True, force creates the environment by deleting the old one.
            album_api_version:
                The api version (runner version) in the environment.

        Raises:
            RuntimeError:
                When the environment could not be created due to whatever reasons.

        """
        env_exists = self.environment_exists(environment_name)
        if force and env_exists:
            self.remove_environment(environment_name)
        else:
            if env_exists:
                raise EnvironmentError(
                    "Environment with name %s already exists!" % environment_name
                )

        env_content = {
            "channels": ["defaults"],
            "dependencies": [
                "python=%s" % DefaultValues.default_solution_python_version.value
            ],
        }

        self._append_runner_and_install(
            album_api_version, environment_name, env_content
        )

    def _append_runner_and_install(
        self, album_api_version, environment_name, environment_content
    ):

        if not album_api_version:
            album_api_version = DefaultValues.runner_api_packet_version.value

        environment_content = self._append_framework_to_yml(
            environment_content, album_api_version
        )
        env_prefix = os.path.normpath(self._environment_name_to_path(environment_name))
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".yml"
        ) as env_file:

            env_file.write(yaml.safe_dump(environment_content))
        subprocess_args = [
            self._get_install_environment_executable(),
            "env",
            "create",
            "--force",
            "--file",
            env_file.name,
            "-p",
            env_prefix,
        ]
        try:
            subcommand.run(subprocess_args, log_output=True)
        except RuntimeError as e:
            # cleanup after failed installation
            if self.environment_exists(environment_name):
                module_logger().debug("Cleanup failed environment creation...")
                self.remove_environment(environment_name)
            raise RuntimeError("Command failed due to reasons above!") from e
        finally:
            os.remove(env_file.name)

    def run_script(self, environment_path, script_full_path, pipe_output=True):
        """Runs a script in the given environment.

        Args:
            environment_path:
                The prefix path of the environment to install the package to.
            script_full_path:
                The full path to the script to run.
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.

        """
        if sys.platform == "win32" or sys.platform == "cygwin":
            # NOTE: WHEN USING 'CONDA RUN' THE CORRECT ENVIRONMENT GETS TEMPORARY ACTIVATED,
            # BUT THE PATH POINTS TO THE WRONG PYTHON (conda base folder python) BECAUSE THE CONDA BASE PATH
            # COMES FIRST IN ENVIRONMENT VARIABLE "%PATH%". THUS, FULL PATH IS NECESSARY TO CALL
            # THE CORRECT PYTHON OR PIP! ToDo: keep track of this!
            subprocess_args = [
                self._conda_executable,
                "run",
                "--no-capture-output",
                "--prefix",
                os.path.normpath(environment_path),
                os.path.normpath(Path(environment_path).joinpath("python")),
                os.path.normpath(script_full_path),
            ]
        else:
            subprocess_args = [
                self._conda_executable,
                "run",
                "--no-capture-output",
                "--prefix",
                os.path.normpath(environment_path),
                "python",
                "-u",
                os.path.normpath(script_full_path),
            ]
        subcommand.run(subprocess_args, pipe_output=pipe_output)

    def set_environment_path(self, environment: IEnvironment):
        path = self.get_environment_path(environment.name())
        module_logger().debug("Set environment path to %s..." % path)
        environment.set_path(path)

    def is_installed(
        self, environment_path: str, package_name, min_package_version=None
    ):
        """Checks if package is installed in a certain version."""
        conda_list = self.list_environment(environment_path)

        for package in conda_list:
            if package["name"] == package_name:
                if min_package_version:
                    if package["version"] == min_package_version:
                        module_logger().debug(
                            "Package %s:%s is installed..."
                            % (package_name, min_package_version)
                        )
                        return True
                    if package["version"] < min_package_version:
                        module_logger().debug(
                            "Package %s:%s is installed. Requirements not set! Reinstalling..."
                            % (package_name, package["version"])
                        )
                        return False
                    if package["version"] > min_package_version:
                        module_logger().debug(
                            "Package %s:%s is installed. Version should be compatible..."
                            % (package_name, package["version"])
                        )
                        return True
                else:
                    module_logger().debug(
                        "Package %s:%s is installed..."
                        % (package_name, package["version"])
                    )
                    return True

        return False

    def run_scripts(self, environment: IEnvironment, scripts, pipe_output=True):
        """Runs the solution in the target environment

        Args:
            scripts:
                List of he scripts calling the solution(s)
            environment:
                The virtual environment used to run the scripts
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.
        """
        if not environment.path():
            raise EnvironmentError(
                "Could not find environment %s. Is the solution installed?"
                % environment.name()
            )

        module_logger().debug("run_in_environment: %s..." % str(environment.path()))

        # first write scripts to disc, create meta-script to execute them in the order of the list
        if len(scripts) > 1:
            script = ""
            for s in scripts:
                fp_step = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                fp_step.write(s)
                fp_step.flush()
                os.fsync(fp_step)
                script += "\nexec(open(r'%s').read())\n" % fp_step.name
                fp_step.close()
        else:
            script = scripts[0]

        self.write_and_run_script(environment, script, pipe_output=pipe_output)

    def write_and_run_script(self, environment, script, pipe_output):
        """Use an environment path and a temporary file to store the script
        Args:

            environment:
                The virtual environment used to run the scripts
            script:
                The virtual environment used to run the scripts
            pipe_output:
                Indicates whether to pipe the output of the subprocess or just return it as is.
        """
        if debug_settings():
            fp = open(
                str(self._configuration.tmp_path().joinpath("album_test.py")), "w"
            )
            module_logger().debug(
                "Executable file in: %s..."
                % str(self._configuration.tmp_path().joinpath("album_test.py"))
            )
        else:
            fp = tempfile.NamedTemporaryFile(mode="w+", delete=False)
            module_logger().debug("Executable file in: %s..." % fp.name)
        fp.write(script)
        fp.flush()
        os.fsync(fp)
        fp.close()
        self.run_script(environment.path(), fp.name, pipe_output=pipe_output)
        Path(fp.name).unlink()

    def create_or_update_env(
        self, environment: IEnvironment, album_api_version: str = None
    ):
        """Creates or updates the environment"""
        if self.environment_exists(environment.name()):
            self.update(environment)
        else:
            self.create(environment, album_api_version)

    def update(self, environment: IEnvironment):
        """Updates the environment"""
        module_logger().debug("Skip installing environment %s..." % environment.name())
        pass  # ToDo: implement and change log message

    def create(self, environment: IEnvironment, album_api_version=None):
        """Creates environment a solution runs in."""
        if environment.yaml_file():
            self.create_environment_from_file(
                environment.yaml_file(), environment.name(), album_api_version
            )
        else:
            module_logger().warning(
                "No yaml file specified. Creating Environment without dependencies!"
            )
            self.create_environment(environment.name(), album_api_version)

    def install(self, environment: IEnvironment, album_api_version=None):
        """Creates or updates an an environment and installs album in the target environment."""
        self.create_or_update_env(environment, album_api_version)
        self.set_environment_path(environment)



from abc import ABCMeta, abstractmethod


class ISearchManager:
    """Interface responsible for searching with keywords through all configured catalogs. Solutions must not be installed
    to be findable in a search request.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def search(self, keywords):
        """Function corresponding to the `search` subcommand of `album`.

        Searches through album catalogs to find closest matching solution.

        """
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from pathlib import Path

from album.runner.core.api.model.coordinates import ICoordinates


class IConfiguration:
    """Configuration of the album framework installation instance.

    This interface manages the cache paths of the album framework installation instance.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def base_cache_path(self):
        """The base path all other cache folder have as parent folder."""
        raise NotImplementedError

    @abstractmethod
    def conda_executable(self):
        """The conda executable. Either a full path to a conda executable/binary or a command"""
        raise NotImplementedError

    @abstractmethod
    def mamba_executable(self):
        """The mamba executable. Either a full path to a mamba executable/binary or a command"""
        raise NotImplementedError

    @abstractmethod
    def cache_path_download(self):
        """Path for solution unspecific downloads."""
        raise NotImplementedError

    @abstractmethod
    def tmp_path(self) -> Path:
        """Path for solution unspecific temporary files of album."""
        raise NotImplementedError

    @abstractmethod
    def environments_path(self) -> Path:
        """Path for solution environments."""
        raise NotImplementedError

    @abstractmethod
    def installation_path(self) -> Path:
        """Path for solution installation files."""
        raise NotImplementedError

    @abstractmethod
    def lnk_path(self) -> Path:
        """Path for targets of internal links, used to shorten paths (thank you Windows)."""
        raise NotImplementedError

    @abstractmethod
    def is_setup(self):
        """If configuration was already initialized."""
        raise NotImplementedError

    @abstractmethod
    def setup(self, base_cache_path=None):
        raise NotImplementedError

    @abstractmethod
    def get_solution_path_suffix(self, coordinates: ICoordinates) -> Path:
        """Returns the suffix path for a solution giving its group, name and version"""
        raise NotImplementedError

    @abstractmethod
    def get_solution_path_suffix_unversioned(self, coordinates: ICoordinates) -> Path:
        """Returns the suffix path for a solution giving its group and name"""
        raise NotImplementedError

    @abstractmethod
    def get_cache_path_catalog(self, catalog_name):
        """Get the cache path to the catalog with a certain ID. Catalog independent!

        Args:
            catalog_name: The ID of the album catalog

        Returns: Path to local cache of a catalog identified by the ID

        """
        raise NotImplementedError

    @abstractmethod
    def get_catalog_collection_path(self):
        """Returns the path of the collection database file."""
        raise NotImplementedError

    @abstractmethod
    def get_catalog_collection_meta_dict(self):
        """Returns the metadata of the collection as a dict."""
        raise NotImplementedError

    @abstractmethod
    def get_catalog_collection_meta_path(self):
        raise NotImplementedError

    @abstractmethod
    def get_initial_catalogs(self):
        """Returns the catalogs initially added to the collection as a dict."""
        raise NotImplementedError

    @abstractmethod
    def get_initial_catalogs_branch_name(self):
        """Returns the default catalogs branches to use."""
        raise NotImplementedError



import os
from enum import Enum
from pathlib import Path

import album.core


class DefaultValues(Enum):
    """Add an entry here to initialize default attributes for a album framework installation instance."""

    album_cite_doi = "arXiv:2110.00601"  # album DOI
    album_cite_text = "Albrecht, Schmidt, Harrington. Album: a framework for scientific data processing with software solutions of heterogeneous tools."  # album text
    album_cite_url = "https://album.solutions"  # album url

    # runner
    runner_api_packet_version = "0.5.1"  # set to None to use with url
    runner_api_packet_name = "album-runner"  # can also point to zip/url like: https://gitlab.com/album-app/album-runner/-/archive/main/album-runner-main.zip
    runner_pip_version = "pip=21.0"
    first_runner_conda_version = "0.5.1"
    default_solution_python_version = "3.9"

    # templates
    catalog_template_url = "https://gitlab.com/album-app/catalogs/templates"  # base URL of available catalog templates

    # deployment & cloning
    catalog_git_user = "album"  # username used for initial push to a new catalog
    catalog_git_email = (
        album.core.__email__
    )  # email used for initial push to a new catalog

    # catalog
    cache_catalog_name = (
        "cache_catalog"  # the default name of the cache catalog (always configured)
    )
    _catalog_url = "https://gitlab.com/album-app/catalogs/default"
    _catalog_branch = "main"
    default_catalog_src = os.getenv(
        "ALBUM_DEFAULT_CATALOG", _catalog_url
    )  # default catalog, either catalog_url or env. variable
    default_catalog_src_branch = os.getenv(
        "ALBUM_DEFAULT_CATALOG_BRANCH", _catalog_branch
    )  # default catalog branch either _catalog_branch or env. variable
    default_catalog_name = (
        "default"  # default catalog, either catalog_url or env. variable
    )
    catalog_collection_name = "album_collection"  # the default name of the Collection
    catalog_collection_db_name = (
        "catalog_collection.db"  # the default name of the Collection DB
    )
    catalog_collection_json_name = (
        "catalog_collection.json"  # the default name of the Collection JSON
    )
    catalog_index_file_name = (
        "album_catalog_index.db"  # the default index file name of the catalog_index
    )
    catalog_index_metafile_json = (
        "album_catalog_index.json"  # the default meta file name of the catalog_index
    )
    catalog_solution_list_file_name = "album_solution_list.json"  # the default file name for exporting the list of solutions of a catalog
    catalog_folder_prefix = (
        "catalogs"  # base folder prefix where all not local catalogs live
    )
    installation_folder_prefix = (
        "installations"  # base folder prefix where installations live
    )
    cache_path_tmp_prefix = "tmp"  # base folder prefix where solution unspecific internal temporary files live
    link_folder_prefix = (
        "lnk"  # base folder prefix where all internal link destinations live
    )
    catalog_solutions_prefix = "solutions"  # base folder prefix where solutions live
    cache_path_download_prefix = "downloads"  # base folder prefix where downloads live
    cache_path_envs_prefix = "envs"  # base folder prefix where environments live in

    # solutions
    solution_default_name = (
        "solution.py"  # default name how solution.py files are called
    )
    solution_yml_default_name = "solution.yml"  # default name of the solution.yml file
    solution_zip_default_name = "solution.zip"  # default name of the solution.zip file
    changelog_default_name = "CHANGELOG.md"  # default name of the changelog file

    # lnk folder prefixes
    lnk_package_prefix = (
        "pck"  # short name of the folder linked to from the catalogs directory
    )
    lnk_env_prefix = (
        "env"  # short name of the folder linked to from the environments directory
    )
    lnk_solution_prefix = (
        "inst"  # short name of the folder linked to from the installations
    )

    # solution specific files living in the installations folder
    solution_app_prefix = "app"  # solution specific app files
    solution_data_prefix = "data"  # solution specific data files
    solution_internal_cache_prefix = (
        "icache"  # solution specific album internal cache files
    )
    solution_user_cache_prefix = (
        "ucache"  # solution specific user cache files, accessible via runner API
    )

    # environment
    default_environment = (
        "album"  # default environment name the album framework operates from
    )

    # album
    app_data_dir = Path.home().joinpath(".album")  # base data path

    # conda
    conda_default_executable = "conda"  # default conda executable
    conda_path = os.getenv(
        "ALBUM_CONDA_PATH", conda_default_executable
    )  # default conda path, either env. var or conda

    # events
    before_run_event_name = "before-run"
    after_run_event_name = "after-run"

    # server
    server_port = 5476  # default port used to launch server
    server_host = "127.0.0.1"  # default host used to launch server. Set to 0.0.0.0 when used in docker



from abc import ABCMeta, abstractmethod


class IEnvironment:
    """Interface managing an environment a solution lives in.

    Each solution lives in its own environment having different dependencies. These can be libraries, programs, etc.
    Each environment has its own environment path and is identified by such. Each album environment has to have
    the album-runner installed for the album framework to be able to run the solution in its environment.
    An environment can be set up by environment file or only by name.

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def name(self):
        raise NotImplementedError

    @abstractmethod
    def cache_path(self):
        raise NotImplementedError

    @abstractmethod
    def yaml_file(self):
        raise NotImplementedError

    @abstractmethod
    def path(self):
        raise NotImplementedError

    @abstractmethod
    def set_path(self, path):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from typing import Optional

from album.runner.core.api.model.solution import ISolution


class IStateManager:
    __metaclass__ = ABCMeta

    @abstractmethod
    def load(self, path) -> Optional[ISolution]:
        raise NotImplementedError



import os
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory

from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.controller.controller import IAlbumController
from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_updates import ISolutionChange, ChangeType
from album.core.api.model.collection_index import ICollectionIndex
from album.core.model.collection_index import CollectionIndex
from album.core.model.default_values import DefaultValues
from album.core.model.link import Link
from album.core.utils.operations.file_operations import (
    copy_folder,
    copy,
    construct_cache_link_target,
    force_remove,
)
from album.core.utils.operations.git_operations import (
    clone_repository_sparse,
    checkout_files,
)
from album.core.utils.operations.resolve_operations import (
    dict_to_coordinates,
    get_zip_name,
    as_tag,
)
from album.core.utils.operations.solution_operations import get_deploy_dict
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution

module_logger = album_logging.get_active_logger


class SolutionHandler(ISolutionHandler):
    """Handles everything inside the Collection responsible for a solution entry.

    Is NOT responsible for resolving paths as this is part of a catalog.
    """

    def __init__(self, album: IAlbumController):
        self.album = album

    def add_or_replace(self, catalog: ICatalog, active_solution: ISolution, path):
        deploy_dict = get_deploy_dict(active_solution)
        self._get_collection_index().add_or_replace_solution(
            catalog.catalog_id(), active_solution.coordinates(), deploy_dict
        )
        # get the install location
        install_location = self.get_solution_package_path(
            catalog, dict_to_coordinates(deploy_dict)
        )

        if Path(path).is_dir():
            copy_folder(path, install_location, copy_root_folder=False)
        else:
            copy(
                path,
                install_location.joinpath(DefaultValues.solution_default_name.value),
            )

    def add_to_cache_catalog(self, active_solution: ISolution, path):
        self.add_or_replace(
            self.album.catalogs().get_cache_catalog(), active_solution, path
        )

    def set_parent(
        self,
        parent_entry: ICollectionIndex.ICollectionSolution,
        child_entry: ICollectionIndex.ICollectionSolution,
    ):

        self._get_collection_index().insert_collection_collection(
            parent_entry.internal()["collection_id"],
            child_entry.internal()["collection_id"],
            parent_entry.internal()["catalog_id"],
            child_entry.internal()["catalog_id"],
        )

    def remove_parent(self, catalog: ICatalog, coordinates: ICoordinates):
        entry = self._get_collection_index().get_solution_by_catalog_grp_name_version(
            catalog.catalog_id(), coordinates, close=False
        )
        self._get_collection_index().remove_parent(entry.internal()["collection_id"])

    def remove_solution(self, catalog: ICatalog, coordinates: ICoordinates):
        self._get_collection_index().remove_solution(catalog.catalog_id(), coordinates)

    def update_solution(
        self, catalog: ICatalog, coordinates: ICoordinates, attrs: dict
    ):
        self._get_collection_index().update_solution(
            catalog.catalog_id(),
            coordinates,
            attrs,
            CollectionIndex.get_collection_column_keys(),
        )

    def apply_change(self, catalog: ICatalog, change: ISolutionChange, override: bool):
        # FIXME handle other tables (tags etc)
        if change.change_type() is ChangeType.ADDED:
            self._get_collection_index().add_or_replace_solution(
                catalog.catalog_id(),
                change.coordinates(),
                catalog.index().get_solution_by_coordinates(change.coordinates()),
            )

        elif change.change_type() is ChangeType.REMOVED:
            self.remove_solution(catalog, change.coordinates())

        elif change.change_type() is ChangeType.CHANGED:
            # get install status before applying change
            installed = change.solution_status()["installed"]
            self._get_collection_index().add_or_replace_solution(
                catalog.catalog_id(),
                change.coordinates(),
                catalog.index().get_solution_by_coordinates(change.coordinates()),
            )
            if installed:
                # set old (install) status and parents again
                self._set_old_db_stat(catalog, change)

            if override and not catalog.is_cache() and installed:
                module_logger().warning(
                    'CAUTION: Solution "%s" seems to be installed.'
                    " The performed operation can leave a broken installation behind "
                    "if dependencies got changed! Consider reinstalling the solution!"
                    % str(change.coordinates())
                )
                self.retrieve_solution(catalog, change.coordinates())

    def _set_old_db_stat(self, catalog, change):
        db_stat = self._get_db_status_dict(change.solution_status())
        self.update_solution(catalog, change.coordinates(), db_stat)
        if change.solution_status()["parent"]:
            self.set_parent(
                change.solution_status()["parent"],
                self._get_collection_index().get_solution_by_catalog_grp_name_version(
                    catalog.catalog_id(), change.coordinates()
                ),  # the new DB representation after the change
            )

    def set_installed(self, catalog: ICatalog, coordinates: ICoordinates):
        self.update_solution(
            catalog,
            coordinates,
            {
                "installed": 1,
                "installation_unfinished": 0,
                "install_date": datetime.now().isoformat(),
            },
        )

    def set_uninstalled(self, catalog: ICatalog, coordinates: ICoordinates):
        self.update_solution(
            catalog, coordinates, {"installed": 0, "installation_unfinished": 0}
        )

    def set_installation_unfinished(self, catalog: ICatalog, coordinates: ICoordinates):
        self.update_solution(
            catalog, coordinates, {"installed": 0, "installation_unfinished": 1}
        )

    def is_installed(self, catalog: ICatalog, coordinates: ICoordinates) -> bool:
        try:
            return self._get_collection_index().is_installed(
                catalog.catalog_id(), coordinates
            )
        except LookupError:
            return False

    def get_solution_package_path(
        self, catalog: ICatalog, coordinates: ICoordinates
    ) -> Link:
        base_link = catalog.path().joinpath(
            self.album.configuration().get_solution_path_suffix(coordinates)
        )
        link_target = construct_cache_link_target(
            self.album.configuration().lnk_path(),
            base_link,
            DefaultValues.lnk_package_prefix.value,
        )
        return Link(link_target).set_link(link=base_link)

    def get_solution_installation_path(
        self, catalog: ICatalog, coordinates: ICoordinates
    ) -> Link:
        installation_path = self.album.configuration().installation_path()
        base_link = installation_path.joinpath(
            catalog.name(),
            coordinates.group(),
            coordinates.name(),
            coordinates.version(),
        )
        link_target = construct_cache_link_target(
            self.album.configuration().lnk_path(),
            base_link,
            DefaultValues.lnk_solution_prefix.value,
        )
        return Link(link_target).set_link(link=base_link)

    def get_solution_file(self, catalog: ICatalog, coordinates: ICoordinates):
        return self.get_solution_package_path(catalog, coordinates).joinpath(
            DefaultValues.solution_default_name.value
        )

    def get_solution_zip_suffix(self, coordinates: ICoordinates):
        return Path("").joinpath(
            self.album.configuration().get_solution_path_suffix(coordinates),
            get_zip_name(coordinates),
        )

    def retrieve_solution(self, catalog: ICatalog, coordinates: ICoordinates):
        if catalog.is_cache():  # no src to download form or src to copy from
            raise RuntimeError("Cannot download from a cache catalog!")
        else:  # src to download from
            solution_download_target = self.get_solution_package_path(
                catalog, coordinates
            )
            self._download_solution(
                catalog.src(), coordinates, solution_download_target
            )

        solution_path = solution_download_target.joinpath(
            DefaultValues.solution_default_name.value
        )

        return solution_path

    def _download_solution(self, src, coordinates: ICoordinates, target: Path):
        solution_pck = str(
            self.album.configuration().get_solution_path_suffix_unversioned(coordinates)
        )
        tag = as_tag(coordinates)
        with TemporaryDirectory(dir=self.album.configuration().tmp_path()) as tmp_dir:
            repo_dir = Path(tmp_dir).joinpath("repo")
            try:
                with clone_repository_sparse(src, tag, repo_dir) as repo:
                    checkout_files(repo, [solution_pck])
                tmp_solution = repo_dir.joinpath(solution_pck)
                if target.exists():
                    for f in os.listdir(str(target)):
                        force_remove(os.path.join(str(target), f))
                copy_folder(tmp_solution, target, copy_root_folder=False)
            finally:
                force_remove(repo_dir)

    def set_cache_paths(self, solution: ISolution, catalog: ICatalog):
        package_path = self.get_solution_package_path(catalog, solution.coordinates())
        solution_path = self.get_solution_installation_path(
            catalog, solution.coordinates()
        )
        solution.installation().set_package_path(package_path)
        solution.installation().set_data_path(
            solution_path.joinpath(DefaultValues.solution_data_prefix.value)
        )
        solution.installation().set_app_path(
            solution_path.joinpath(DefaultValues.solution_app_prefix.value)
        )
        solution.installation().set_internal_cache_path(
            solution_path.joinpath(DefaultValues.solution_internal_cache_prefix.value)
        )
        solution.installation().set_user_cache_path(
            solution_path.joinpath(DefaultValues.solution_user_cache_prefix.value)
        )

    def _get_collection_index(self):
        return self.album.collection_manager().get_collection_index()

    @staticmethod
    def _get_db_status_dict(internal_status: dict) -> dict:
        """Everything that should NOT change internally when an UPDATE change is performed on a solution."""
        r = {
            "installed": internal_status["installed"],
            "install_date": internal_status["install_date"],
            "last_execution": internal_status["last_execution"],
        }

        return r



from queue import Queue

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.test_manager import ITestManager
from album.runner import album_logging
from album.runner.core.model.script_creator import ScriptCreatorTest

module_logger = album_logging.get_active_logger


class TestManager(ITestManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def test(self, solution_to_resolve: str, args=None):
        if args is None:
            args = [""]

        resolve_result = self.album.collection_manager().resolve_installed_and_load(
            solution_to_resolve
        )
        solution = resolve_result.loaded_solution()

        if (
            solution.setup().pre_test
            and callable(solution.setup().pre_test)
            and solution.setup().test
            and callable(solution.setup().test)
        ) or (
            not solution.setup().pre_test
            and solution.setup().test
            and callable(solution.setup().test)
        ):
            queue = Queue()
            script_test_creator = ScriptCreatorTest()

            # do not run queue immediately
            self.album.script_manager().build_queue(
                resolve_result, queue, script_test_creator, False, args
            )

            # runs the queue
            self.album.script_manager().run_queue(queue)

            module_logger().info(
                'Ran test routine for "%s"!'
                % resolve_result.loaded_solution().coordinates().name()
            )
        else:
            module_logger().warning(
                'No "test" routine configured for solution "%s"! Skipping...'
                % solution.coordinates().name()
            )



from abc import ABCMeta, abstractmethod

from album.core.api.model.task import ITask


class ITaskManager:
    """Interface for retrieving the status of a solution."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def get_task(self, task_id) -> ITask:
        """Get a task managed by the task manager."""
        raise NotImplementedError

    @abstractmethod
    def get_status(self, task) -> ITask.Status:
        """Get the status of a task managed by the task manager."""
        raise NotImplementedError

    @abstractmethod
    def create_and_register_task(self, method, args) -> str:
        raise NotImplementedError

    @abstractmethod
    def register_task(self, task: ITask) -> str:
        raise NotImplementedError

    @abstractmethod
    def finish_tasks(self):
        raise NotImplementedError



import pkgutil
from pathlib import Path

import album
from album.core.utils.operations.file_operations import copy_in_file
from album.core.utils.operations.resolve_operations import get_zip_name
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.solution import ISolution


def create_docker_file(active_solution: ISolution, target_folder: Path) -> Path:
    """Uses the template to create a docker file for the solution which gets deployed.

    Returns:
        The path to the docker file.
    """
    coordinates = active_solution.coordinates()
    zip_name = get_zip_name(coordinates)

    docker_file = target_folder.joinpath("Dockerfile")

    docker_file_stream = pkgutil.get_data(
        "album.docker", "Dockerfile_solution_template.txt"
    ).decode()

    docker_file_stream = docker_file_stream.replace("<version>", album.core.__version__)
    docker_file_stream = docker_file_stream.replace("<name>", zip_name)
    docker_file_stream = docker_file_stream.replace("<run_name>", str(coordinates))
    author = (
        "; ".join(active_solution.setup().solution_creators)
        if active_solution.setup().solution_creators
        else '""'
    )
    docker_file_stream = docker_file_stream.replace("<maintainer>", author)

    # replace template with entries and copy dockerfile to deploy_src
    get_active_logger().debug("Writing docker file to: %s..." % str(docker_file))
    copy_in_file(docker_file_stream, docker_file)

    return docker_file



from typing import Optional

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.install_manager import IInstallManager
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.api.model.environment import IEnvironment
from album.core.controller.environment_manager import EnvironmentManager
from album.core.model.resolve_result import ResolveResult
from album.core.utils.operations.file_operations import remove_link
from album.core.utils.operations.resolve_operations import (
    clean_resolve_tmp,
    build_resolve_string,
    dict_to_coordinates,
)
from album.core.utils.operations.solution_operations import (
    get_deploy_dict,
    get_parent_dict,
)
from album.core.utils.operations.view_operations import get_solution_run_call_as_string
from album.runner import album_logging
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.script_creator import (
    ScriptCreatorInstall,
    ScriptCreatorUnInstall,
)

module_logger = album_logging.get_active_logger


class InstallManager(IInstallManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def install(self, solution_to_resolve: str, argv=None):
        # this needs to happen before any (potentially not completely installed) solution is resolved
        self.clean_unfinished_installations()

        resolve_result = self.album.collection_manager().resolve_and_load(
            solution_to_resolve
        )
        self._install_resolve_result(resolve_result, argv, parent=False)

    def _resolve_result_is_installed(self, resolve_result: ICollectionSolution) -> bool:
        """Checks whether a resolve_result is already installed."""
        if resolve_result.database_entry():  # we know the solution is in the collection
            return (
                self.album.collection_manager()
                .get_collection_index()
                .is_installed(
                    resolve_result.catalog().catalog_id(), resolve_result.coordinates()
                )
            )
        return False

    def _install_resolve_result(
        self, resolve_result: ICollectionSolution, argv=None, parent=False
    ):
        """Installs a resolve result.

        CAUTION: Solution must be loaded!
        """
        # Load solution
        if not resolve_result.catalog():
            raise RuntimeError(
                "Solution cannot be installed without being associated with a catalog!"
            )
        elif not parent and self._resolve_result_is_installed(resolve_result):
            raise RuntimeError("Solution already installed. Uninstall solution first!")
        elif parent and self._resolve_result_is_installed(resolve_result):
            return resolve_result  # solution already installed
        else:
            module_logger().debug(
                'solution loaded from catalog "%s": %s...'
                % (
                    resolve_result.catalog().name(),
                    str(resolve_result.loaded_solution().coordinates()),
                )
            )
        if not parent:
            module_logger().info(
                'Installing "%s"...'
                % resolve_result.loaded_solution().coordinates().name()
            )
            # fail when already installed
            if self._resolve_result_is_installed(resolve_result):
                raise RuntimeError(
                    "Solution already installed. Uninstall solution first!"
                )
        else:
            module_logger().info(
                'Installing parent solution "%s"...'
                % resolve_result.loaded_solution().coordinates().name()
            )

        self._register(resolve_result)

        if not parent:
            if resolve_result.catalog().is_cache():
                # always clean after registration to a catalog!
                clean_resolve_tmp(self.album.configuration().tmp_path())

        # mark as "installation unfinished"
        self.album.solutions().set_installation_unfinished(
            resolve_result.catalog(), resolve_result.coordinates()
        )

        # run installation recursively
        self._install_active_solution(resolve_result, argv)

        # mark as installed and remove "installation unfinished"
        self.album.solutions().set_installed(
            resolve_result.catalog(), resolve_result.coordinates()
        )
        if parent:
            module_logger().info(
                "Installed parent solution %s!" % resolve_result.coordinates().name()
            )
        else:
            module_logger().info(
                "Installed %s! Learn more about the solution by calling `album info %s`."
                % (resolve_result.coordinates().name(), resolve_result.coordinates())
            )

    def _register(self, resolve_result: ICollectionSolution):
        """Registers a resolve result in the collection"""
        # register in collection
        if resolve_result.catalog().is_cache():
            # a cache catalog is living in the collection so no need to update, we can add it directly
            self.album.solutions().add_to_cache_catalog(
                resolve_result.loaded_solution(),
                resolve_result.path().parent,  # the directory holding the solution file
            )
        else:
            # update the collection holding the solution entry
            self._update_in_collection_index(resolve_result)
        resolve_result.set_database_entry(
            self.album.collection_manager()
            .get_collection_index()
            .get_solution_by_catalog_grp_name_version(
                resolve_result.catalog().catalog_id(), resolve_result.coordinates()
            )
        )

    def _remove_parent(self, resolve_result: ICollectionSolution):
        """Sets the parent of a solution"""
        self.album.solutions().remove_parent(
            resolve_result.catalog(), resolve_result.coordinates()
        )

    def _update_in_collection_index(self, resolve_result: ICollectionSolution):
        """Updates the collection entry of the resolved solution"""
        # update the solution in the collection
        self.album.solutions().update_solution(
            resolve_result.catalog(),
            resolve_result.coordinates(),
            get_deploy_dict(resolve_result.loaded_solution()),
        )

    def _install_active_solution(
        self, collection_solution: ICollectionSolution, argv=None
    ) -> Optional[ICollectionSolution]:
        """Installation routine for a loaded solution."""
        # install environment
        if argv is None:
            argv = [""]

        parent_resolve_result = None

        parent = get_parent_dict(collection_solution.loaded_solution())
        if parent:
            # install dependencies first. Recursive call to install with dependencies
            try:
                parent_resolve_result = self._install_parent(parent)
            except Exception as e:
                module_logger().error("Exception when installing parent:")
                raise e

            if parent_resolve_result:
                self.album.solutions().set_parent(
                    parent_resolve_result.database_entry(),
                    collection_solution.database_entry(),
                )
            else:
                self._remove_parent(collection_solution)

            collection_solution.set_database_entry(
                self.album.collection_manager()
                .get_collection_index()
                .get_solution_by_catalog_grp_name_version(
                    collection_solution.catalog().catalog_id(),
                    collection_solution.coordinates(),
                )
            )

            # resolve environment - at this point all parents should be already installed
            environment = self.album.environment_manager().set_environment(
                collection_solution
            )
        else:
            environment = self.album.environment_manager().install_environment(
                collection_solution
            )

        self._run_solution_install_routine(
            collection_solution.loaded_solution(), environment, argv
        )

        return parent_resolve_result

    def _run_solution_install_routine(
        self, active_solution: ISolution, environment: IEnvironment, argv
    ):
        """Run install routine of album if specified"""
        script_creator_install = ScriptCreatorInstall()

        if active_solution.setup().install and callable(
            active_solution.setup().install
        ):
            module_logger().debug("Creating install script...")
            script = script_creator_install.create_script(active_solution, argv)
            module_logger().debug("Calling install routine specified in solution...")
            self.album.environment_manager().run_scripts(environment, [script])
        else:
            module_logger().debug(
                'No "install" routine configured for solution "%s". Skipping...'
                % active_solution.coordinates().name()
            )

    def _install_parent(self, parent_dict: dict) -> ICollectionSolution:
        """Installs a parent of a solution given its description as a dictionary."""
        resolve_solution = build_resolve_string(parent_dict)
        resolve_result = self.album.collection_manager().resolve_and_load(
            resolve_solution
        )

        # recursive installation call. Not failing for already installed solutions. parent set to "True"
        self._install_resolve_result(resolve_result, parent=True)
        return resolve_result

    def uninstall(self, solution_to_resolve: str, rm_dep=False, argv=None):
        """Internal installation entry point for `uninstall` subcommand of `album`."""

        resolve_result = self.album.collection_manager().resolve_installed_and_load(
            solution_to_resolve
        )

        module_logger().info(
            'Uninstalling "%s"...' % resolve_result.coordinates().name()
        )

        if argv is None:
            argv = [""]

        parent = resolve_result.database_entry().internal()["parent"]
        # get the environment
        environment = None
        try:
            environment = self.album.environment_manager().set_environment(
                resolve_result
            )
            self._run_solution_uninstall_routine(
                resolve_result.loaded_solution(), environment, argv
            )

            if not parent:
                self.album.environment_manager().remove_environment(environment)

        except LookupError:
            # environment might have been deleted manually
            pass
        finally:
            if environment and not parent:
                EnvironmentManager.remove_disc_content_from_environment(environment)

        if resolve_result.database_entry().internal()["children"]:
            children = []
            for dependency_dict in resolve_result.database_entry().internal()[
                "children"
            ]:
                # get the child entry
                child_solution = (
                    self.album.collection_manager()
                    .get_collection_index()
                    .get_solution_by_collection_id(
                        dependency_dict["collection_id_child"]
                    )
                )
                if child_solution.internal()["installed"]:
                    children.append(str(dict_to_coordinates(child_solution.setup())))

            if children:
                module_logger().info(
                    "The following solutions depend on this installation: %s. Not uninstalling %s..."
                    % (", ".join(children), str(resolve_result.coordinates()))
                )
                if parent:
                    return

                raise RuntimeError(
                    'Cannot uninstall "%s". Other solution depend on this installation! '
                    "Inspect log for more information!" % resolve_result.coordinates()
                )

        self._remove_disc_content_from_solution(resolve_result)

        if resolve_result.catalog().is_cache():
            self.album.solutions().remove_solution(
                resolve_result.catalog(), resolve_result.coordinates()
            )
        else:
            self.album.solutions().set_uninstalled(
                resolve_result.catalog(), resolve_result.coordinates()
            )

        if rm_dep:  # remove dependencies (parent of the solution) last
            self._remove_dependencies(resolve_result.loaded_solution(), rm_dep)

        module_logger().info('Uninstalled "%s"!' % resolve_result.coordinates().name())

    def _run_solution_uninstall_routine(
        self, active_solution: ISolution, environment: IEnvironment, argv
    ):
        """Run uninstall routine of album if specified. Expects environment to be set!"""
        script_creator_un_install = ScriptCreatorUnInstall()

        if active_solution.setup().uninstall and callable(
            active_solution.setup().uninstall
        ):
            module_logger().debug("Creating uninstall script...")
            script = script_creator_un_install.create_script(active_solution, argv)
            module_logger().debug("Calling uninstall routine specified in solution...")
            album_logging.configure_logging("uninstall")
            self.album.environment_manager().run_scripts(environment, [script])
            album_logging.pop_active_logger()
        else:
            module_logger().debug(
                'No "uninstall" routine configured for solution "%s"! Skipping...'
                % active_solution.coordinates().name()
            )

    def _remove_dependencies(self, solution: ISolution, rm_dep=False):
        parent = get_parent_dict(solution)
        if parent:
            # recursive call to remove the parent
            resolve_solution = build_resolve_string(parent)
            self.uninstall(resolve_solution, rm_dep)

    def clean_unfinished_installations(self):
        collection_solution_list = (
            self.album.collection_manager()
            .get_collection_index()
            .get_unfinished_installation_solutions()
        )
        for collection_solution in collection_solution_list:
            catalog = self.album.catalogs().get_by_id(
                collection_solution.internal()["catalog_id"]
            )
            path = self.album.solutions().get_solution_file(
                catalog, dict_to_coordinates(collection_solution.setup())
            )
            coordinates = dict_to_coordinates(collection_solution.setup())

            resolve = ResolveResult(
                path=path,
                catalog=catalog,
                collection_entry=collection_solution,
                coordinates=coordinates,
            )

            self.album.collection_manager().retrieve_and_load_resolve_result(resolve)

            self.album.solutions().set_cache_paths(
                resolve.loaded_solution(), resolve.catalog()
            )

            # only remove environment when solution has its own environment
            if not get_parent_dict(resolve.loaded_solution()):
                self._clean_unfinished_installations_environment(resolve)

            self._remove_disc_content_from_solution(resolve)

            if resolve.catalog().is_cache():
                self.album.solutions().remove_solution(resolve.catalog(), coordinates)
            else:
                self.album.solutions().set_uninstalled(resolve.catalog(), coordinates)

    def _clean_unfinished_installations_environment(self, resolve: ICollectionSolution):
        try:
            environment = self.album.environment_manager().set_environment(resolve)
            self.album.environment_manager().remove_environment(environment)
        except LookupError:
            pass

    def _remove_disc_content_from_solution(self, resolve_result: ICollectionSolution):
        remove_link(
            self.album.collection_manager()
            .solutions()
            .get_solution_installation_path(
                resolve_result.catalog(), resolve_result.coordinates()
            )
        )
        remove_link(
            self.album.collection_manager()
            .solutions()
            .get_solution_package_path(
                resolve_result.catalog(), resolve_result.coordinates()
            )
        )



import os
import platform
import shutil
import sys
from pathlib import Path

from album.core.api.model.configuration import IConfiguration
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import (
    create_paths_recursively,
    force_remove,
    get_dict_from_json,
)
from album.runner import album_logging
from album.runner.core.model.coordinates import Coordinates

module_logger = album_logging.get_active_logger


class Configuration(IConfiguration):
    def __init__(self):
        self._is_setup = False
        self._base_cache_path = None
        self._conda_executable = None
        self._mamba_executable = None
        self._tmp_path = None
        self._cache_path_envs = None
        self._catalog_collection_path = None
        self._installation_path = None
        self._lnk_path = None

    def base_cache_path(self):
        return self._base_cache_path

    def conda_executable(self):
        return self._conda_executable

    def mamba_executable(self):
        return self._mamba_executable

    def installation_path(self):
        return self._installation_path

    def cache_path_download(self):
        return self._cache_path_download

    def tmp_path(self):
        return self._tmp_path

    def environments_path(self):
        return self._cache_path_envs

    def lnk_path(self):
        return self._lnk_path

    def is_setup(self):
        return self._is_setup

    def setup(self, base_cache_path=None):
        if self._is_setup:
            raise RuntimeError(
                "Configuration::setup was already called and should not be called twice."
            )
        self._is_setup = True
        # base root path where everything lives
        self._base_cache_path = Path(
            os.getenv("ALBUM_BASE_CACHE_PATH", DefaultValues.app_data_dir.value)
        )
        if base_cache_path:
            self._base_cache_path = Path(base_cache_path)

        # conda executable
        conda_path = DefaultValues.conda_path.value
        operation_system = platform.system().lower()
        if conda_path is not DefaultValues.conda_default_executable.value:
            self._conda_executable = self._build_conda_executable(conda_path)
        else:
            self._conda_executable = conda_path
            if "windows" in operation_system:
                self._conda_executable = shutil.which(self._conda_executable)

        self._mamba_executable = shutil.which("mamba")

        self._cache_path_download = self._base_cache_path.joinpath(
            DefaultValues.cache_path_download_prefix.value
        )
        self._cache_path_envs = self._base_cache_path.joinpath(
            DefaultValues.cache_path_envs_prefix.value
        )
        self._catalog_collection_path = self._base_cache_path.joinpath(
            DefaultValues.catalog_folder_prefix.value
        )
        self._installation_path = self._base_cache_path.joinpath(
            DefaultValues.installation_folder_prefix.value
        )
        self._tmp_path = self._base_cache_path.joinpath(
            DefaultValues.cache_path_tmp_prefix.value
        )
        self._lnk_path = self._base_cache_path.joinpath(
            DefaultValues.link_folder_prefix.value
        )
        self._empty_tmp()
        create_paths_recursively(
            [
                self._tmp_path,
                self._cache_path_download,
                self._cache_path_envs,
                self._catalog_collection_path,
                self._installation_path,
            ]
        )

    @staticmethod
    def _build_conda_executable(conda_path):
        operation_system = sys.platform
        if operation_system == "linux" or operation_system == "darwin":
            return str(Path(conda_path).joinpath("bin", "conda"))
        else:
            return str(Path(conda_path).joinpath("Scripts", "conda.exe"))

    def get_solution_path_suffix(self, coordinates: Coordinates) -> Path:
        return Path("").joinpath(
            DefaultValues.catalog_solutions_prefix.value,
            coordinates.group(),
            coordinates.name(),
            coordinates.version(),
        )

    def get_solution_path_suffix_unversioned(self, coordinates: Coordinates) -> Path:
        return Path("").joinpath(
            DefaultValues.catalog_solutions_prefix.value,
            coordinates.group(),
            coordinates.name(),
        )

    def get_cache_path_catalog(self, catalog_name):
        return self._base_cache_path.joinpath(
            DefaultValues.catalog_folder_prefix.value, catalog_name
        )

    def get_catalog_collection_path(self):
        collection_db_path = Path(self._catalog_collection_path).joinpath(
            DefaultValues.catalog_collection_db_name.value
        )
        return collection_db_path

    def get_catalog_collection_meta_dict(self):
        """Returns the metadata of the collection as a dict."""
        catalog_collection_json = self.get_catalog_collection_meta_path()
        if not catalog_collection_json.exists():
            return None
        catalog_collection_dict = get_dict_from_json(catalog_collection_json)
        return catalog_collection_dict

    def get_catalog_collection_meta_path(self):
        return self._catalog_collection_path.parent.joinpath(
            DefaultValues.catalog_collection_json_name.value
        )

    def get_initial_catalogs(self):
        return {
            DefaultValues.default_catalog_name.value: DefaultValues.default_catalog_src.value
        }

    def get_initial_catalogs_branch_name(self):
        return {
            DefaultValues.default_catalog_name.value: DefaultValues.default_catalog_src_branch.value
        }

    def _empty_tmp(self):
        """Removes the content of the tmp folder"""
        # this should not be done since there could be links in tmp_user or tmp_internal which have to be resolved when deleting them
        # force_remove(self._cache_path_tmp_user)
        # force_remove(self._cache_path_tmp_internal)
        force_remove(self._tmp_path)



from pathlib import Path
from tempfile import TemporaryDirectory

import git.exc
from git import Repo

from album.core.api.controller.clone_manager import ICloneManager
from album.core.api.controller.controller import IAlbumController
from album.core.model.default_values import DefaultValues
from album.core.utils.operations import url_operations
from album.core.utils.operations.file_operations import (
    create_path_recursively,
    get_dict_from_json,
    list_files_recursively,
    force_remove,
    copy_folder,
    unzip_archive,
)
from album.core.utils.operations.git_operations import (
    create_bare_repository,
    clone_repository,
    add_files_commit_and_push,
    checkout_main,
)
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class CloneManager(ICloneManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def clone(
        self,
        path: str,
        target_dir: str,
        name: str,
        git_email: str = None,
        git_name: str = None,
    ) -> None:
        if path.startswith("template:"):
            try:
                self._clone_catalog_template(
                    path[len("template:") :], target_dir, name, git_email, git_name
                )
            except (LookupError, ValueError):
                raise LookupError(
                    "Cannot resolve %s - make sure it's a valid name of a template located in %s!",
                    (path, DefaultValues.catalog_template_url.value),
                )
        else:
            target_path = Path(target_dir).joinpath(name)
            self._clone_solution(path, target_path)

    def _clone_solution(self, path, target_path):
        """Copies a solution (by resolving and downloading) to a given target path."""
        resolve_result = self.album.collection_manager().resolve(path)

        copy_folder(resolve_result.path().parent, target_path, copy_root_folder=False)

        module_logger().info(
            "Copied solution %s to %s!" % (resolve_result.path(), target_path)
        )

    def _clone_catalog_template(
        self,
        template_name,
        target_path,
        catalog_name,
        git_email: str = None,
        git_name: str = None,
    ):
        """Clones a template by looking up the template name in the template catalog"""
        template_url = "%s/%s/-/archive/main/%s-main.zip" % (
            DefaultValues.catalog_template_url.value,
            template_name,
            template_name,
        )
        if url_operations.is_downloadable(template_url):
            download_zip_target = (
                self.album.configuration()
                .cache_path_download()
                .joinpath(template_name + ".zip")
            )
            download_unzip_target = (
                self.album.configuration().cache_path_download().joinpath(template_name)
            )

            url_operations.download_resource(template_url, download_zip_target)
            module_logger().debug(
                "Downloaded template from %s to %s!"
                % (template_url, download_zip_target)
            )

            unzip_archive(download_zip_target, download_unzip_target)
            download_unzip_target_subdir = download_unzip_target.joinpath(
                f"{template_name}-main"
            )

            self.setup_repository_from_template(
                target_path,
                download_unzip_target_subdir,
                catalog_name,
                git_email,
                git_name,
            )

            module_logger().info(
                'Initialized new catalog "%s" from template "%s" in %s!'
                % (catalog_name, template_name, target_path)
            )
            return True
        return False

    def setup_repository_from_template(
        self,
        target_path,
        template_folder,
        catalog_name,
        git_email: str = None,
        git_name: str = None,
    ):
        # test if target_path can already be cloned:

        try:
            with TemporaryDirectory(
                dir=self.album.configuration().tmp_path()
            ) as tmp_dir:
                tmp_clone_path = Path(tmp_dir).joinpath("clone")
                with clone_repository(target_path, tmp_clone_path) as repo:
                    self._copy_template_into_repository(
                        repo, template_folder, catalog_name, git_email, git_name
                    )
                force_remove(tmp_clone_path)
        except git.GitCommandError:
            create_path_recursively(target_path)
            create_bare_repository(target_path)
            with TemporaryDirectory(
                dir=self.album.configuration().tmp_path()
            ) as tmp_dir:
                tmp_clone_path = Path(tmp_dir).joinpath("clone")
                with clone_repository(target_path, tmp_clone_path) as repo:
                    self._copy_template_into_repository(
                        repo,
                        template_folder,
                        catalog_name,
                        email=DefaultValues.catalog_git_email.value,
                        username=DefaultValues.catalog_git_user.value,
                    )
                force_remove(tmp_clone_path)

    def _copy_template_into_repository(
        self, repo: Repo, template_folder, catalog_name: str, email=None, username=None
    ):
        head = checkout_main(repo)

        copy_folder(template_folder, repo.working_tree_dir, copy_root_folder=False)

        # create the metadata in repository clone
        catalog_type = self._get_catalog_type_from_template(template_folder)
        self.album.catalogs().create_new_metadata(
            repo.working_tree_dir, catalog_name, catalog_type
        )

        add_files_commit_and_push(
            head,
            list_files_recursively(repo.working_tree_dir),
            'Setting up "%s" catalog!' % catalog_name,
            push=True,
            force=False,
            email=email,
            username=username,
        )

    @staticmethod
    def _get_catalog_type_from_template(template_base_path):
        template_metadata_path = CloneManager._get_metadata_path_from_template(
            template_base_path
        )

        if template_metadata_path.is_file():
            template_metadata = get_dict_from_json(template_metadata_path)
        else:
            raise FileNotFoundError(
                "Could not find file %s!" % str(template_metadata_path)
            )

        return template_metadata["type"]

    @staticmethod
    def _get_metadata_path_from_template(template_base_path):
        template_base_path = Path(template_base_path)
        template_metadata_path = template_base_path.joinpath(
            DefaultValues.catalog_index_metafile_json.value
        )

        return template_metadata_path
